# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines ADEV primitives."""

import jax
import jax._src.core
import jax._src.dtypes as jax_dtypes
import jax.numpy as jnp
from jax._src.ad_util import Zero
from jax._src.core import (
    get_aval,
    raise_to_shaped,
)
from jax.interpreters.ad import zeros_like_aval
from tensorflow_probability.substrates import jax as tfp

from genjax._src.adev.core import (
    ADEVPrimitive,
    Dual,
    DualTree,
    TailCallADEVPrimitive,
)
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    PRNGKey,
)

tfd = tfp.distributions

# These methods are pulled from jax.interpreters.ad:


def instantiate_zeros(tangent):
    return zeros_like_aval(tangent.aval) if type(tangent) is Zero else tangent


def zeros_like_jaxval(val):
    return zeros_like_aval(raise_to_shaped(get_aval(val)))


def recast_to_float0(primal, tangent):
    if (
        jax._src.core.primal_dtype_to_tangent_dtype(jax_dtypes.dtype(primal))
        == jax_dtypes.float0
    ):
        return Zero(jax._src.core.get_aval(primal).at_least_vspace())
    else:
        return tangent


def zero(v):
    ad_zero = recast_to_float0(v, zeros_like_jaxval(v))
    return instantiate_zeros(ad_zero)


################################
# Gradient strategy primitives #
################################


@Pytree.dataclass
class REINFORCE(ADEVPrimitive):
    sample_function: Callable[..., Any] = Pytree.static()
    differentiable_logpdf: Callable[..., Any] = Pytree.static()

    def sample(self, key, *args):
        return self.sample_function(key, *args)

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (_, kdual) = konts
        primals = Dual.tree_primal(dual_tree)
        tangents = Dual.tree_tangent(dual_tree)
        key, sub_key = jax.random.split(key)
        v = self.sample(sub_key, *primals)
        dual_tree = Dual.tree_pure(v)
        out_dual = kdual(key, dual_tree)
        (out_primal,), (out_tangent,) = Dual.tree_unzip(out_dual)
        _, lp_tangent = jax.jvp(
            self.differentiable_logpdf,
            (v, *primals),
            (zero(v), *tangents),
        )
        return Dual(out_primal, out_tangent + (out_primal * lp_tangent))


def reinforce(sample_func, logpdf_func):
    return REINFORCE(sample_func, logpdf_func)


###########################
# Distribution primitives #
###########################


@Pytree.dataclass
class FlipEnum(ADEVPrimitive):
    def sample(self, key, *args):
        (probs,) = args
        return 1 == tfd.Bernoulli(probs=probs).sample(seed=key)

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (_, kdual) = konts
        (p_primal,) = Dual.tree_primal(dual_tree)
        (p_tangent,) = Dual.tree_tangent(dual_tree)
        true_dual = kdual(
            key,
            Dual(jnp.array(True), jnp.zeros_like(jnp.array(True))),
        )
        false_dual = kdual(
            key,
            Dual(jnp.array(False), jnp.zeros_like(jnp.array(False))),
        )
        (true_primal,), (true_tangent,) = Dual.tree_unzip(true_dual)
        (false_primal,), (false_tangent,) = Dual.tree_unzip(false_dual)

        def _inner(p, tl, fl):
            return p * tl + (1 - p) * fl

        out_primal, out_tangent = jax.jvp(
            _inner,
            (p_primal, true_primal, false_primal),
            (p_tangent, true_tangent, false_tangent),
        )
        return Dual(out_primal, out_tangent)


flip_enum = FlipEnum()


@Pytree.dataclass
class FlipMVD(ADEVPrimitive):
    def sample(self, key, *args):
        p = (args,)
        return 1 == tfd.Bernoulli(probs=p).sample(seed=key)

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (kpure, kdual) = konts
        (p_primal,) = Dual.tree_primal(dual_tree)
        (p_tangent,) = Dual.tree_primal(dual_tree)
        key, sub_key = jax.random.split(key)
        v = tfd.Bernoulli(probs=p_primal).sample(seed=sub_key)
        b = v == 1
        b_primal, b_tangent = kdual(key, (b,), (jnp.zeros_like(b),))
        other = kpure(key, jnp.logical_not(b))
        est = ((-1) ** v) * (other - b_primal)
        return Dual(b_primal, b_tangent + est * p_tangent)


flip_mvd = FlipMVD()


@Pytree.dataclass
class FlipEnumParallel(ADEVPrimitive):
    def sample(self, key, *args):
        (p,) = args
        return 1 == tfd.Bernoulli(probs=p).sample(seed=key)

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (_, kdual) = konts
        (p_primal,) = Dual.tree_primal(dual_tree)
        (p_tangent,) = Dual.tree_tangent(dual_tree)
        sub_keys = jax.random.split(key, 2)
        ret_primals, ret_tangents = jax.vmap(kdual)(
            sub_keys,
            (jnp.array([True, False]),),
            (jnp.zeros_like(jnp.array([True, False]))),
        )

        def _inner(p, ret):
            return jnp.sum(jnp.array([p, 1 - p]) * ret)

        return Dual(
            *jax.jvp(
                _inner,
                (p_primal, ret_primals),
                (p_tangent, ret_tangents),
            )
        )


flip_enum_parallel = FlipEnumParallel()


@Pytree.dataclass
class CategoricalEnumParallel(ADEVPrimitive):
    def sample(self, key, *args):
        (probs,) = args
        return tfd.Categorical(probs=probs).sample(seed=key)

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (_, kdual) = konts
        (probs_primal,) = Dual.tree_primal(dual_tree)
        (probs_tangent,) = Dual.tree_tangent(dual_tree)
        idxs = jnp.arange(len(probs_primal))
        sub_keys = jax.random.split(key, len(probs_primal))
        ret_primals, ret_tangents = jax.vmap(kdual)(
            sub_keys, (idxs,), (jnp.zeros_like(idxs),)
        )

        def _inner(probs, primals):
            return jnp.sum(jax.nn.softmax(probs) * primals)

        return Dual(
            *jax.jvp(
                _inner,
                (probs_primal, ret_primals),
                (probs_tangent, ret_tangents),
            )
        )


categorical_enum_parallel = CategoricalEnumParallel()

flip_reinforce = reinforce(
    lambda key, p: 1 == tfd.Bernoulli(probs=p).sample(seed=key),
    lambda v, p: tfd.Bernoulli(probs=p).log_prob(v),
)

geometric_reinforce = reinforce(
    lambda key, args: tfd.Geometric(*args).sample(seed=key),
    lambda v, args: tfd.Geometric(*args).log_prob(v),
)

normal_reinforce = reinforce(
    lambda key, loc, scale: tfd.Normal(loc, scale).sample(seed=key),
    lambda v, loc, scale: tfd.Normal(loc, scale).log_prob(v),
)


@Pytree.dataclass
class NormalREPARAM(TailCallADEVPrimitive):
    def sample(self, key, *args):
        loc, scale_diag = args
        return tfd.Normal(loc=loc, scale=scale_diag).sample(seed=key)

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
    ) -> Dual:
        (mu_primal, sigma_primal) = Dual.tree_primal(dual_tree)
        (mu_tangent, sigma_tangent) = Dual.tree_tangent(dual_tree)
        key, sub_key = jax.random.split(key)
        eps = tfd.Normal(loc=0.0, scale=1.0).sample(seed=sub_key)

        def _inner(mu, sigma):
            return mu + sigma * eps

        primal_out, tangent_out = jax.jvp(
            _inner,
            (mu_primal, sigma_primal),
            (mu_tangent, sigma_tangent),
        )
        return Dual(primal_out, tangent_out)


normal_reparam = NormalREPARAM()


@Pytree.dataclass
class MvNormalDiagREPARAM(TailCallADEVPrimitive):
    def sample(self, key, *args):
        loc, scale_diag = args
        return tfd.MultivariateNormalDiag(loc=loc, scale_diag=scale_diag).sample(
            seed=key
        )

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
    ):
        (loc_primal, diag_scale_primal) = Dual.tree_primal(dual_tree)
        (loc_tangent, diag_scale_tangent) = Dual.tree_tangent(dual_tree)
        key, sub_key = jax.random.split(key)

        eps = tfd.Normal(loc=0.0, scale=1.0).sample(
            sample_shape=loc_primal.shape, seed=sub_key
        )

        # This takes N samples from N(0.0, 1.0) and transforms
        # them to MvNormalDiag(loc, diag_scale).
        def _inner(loc, diag_scale):
            return loc + jnp.multiply(diag_scale, eps)

        primal_out, tangent_out = jax.jvp(
            _inner,
            (loc_primal, diag_scale_primal),
            (loc_tangent, diag_scale_tangent),
        )
        return Dual(primal_out, tangent_out)


mv_normal_diag_reparam = MvNormalDiagREPARAM()


@Pytree.dataclass
class MvNormalREPARAM(TailCallADEVPrimitive):
    def sample(self, key, *args):
        mu, sigma = args
        v = tfd.MultivariateNormalFullCovariance(
            loc=mu, covariance_matrix=sigma
        ).sample(seed=key)
        return v

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
    ):
        (mu_primal, cov_primal) = Dual.tree_primal(dual_tree)
        (mu_tangent, cov_tangent) = Dual.tree_primal(dual_tree)
        key, sub_key = jax.random.split(key)

        eps = tfd.Normal(loc=0.0, scale=1.0).sample(len(mu_primal), seed=sub_key)

        def _inner(eps, mu, cov):
            L = jnp.linalg.cholesky(cov)
            return mu + L @ eps

        primal_out, tangent_out = jax.jvp(
            _inner,
            (eps, mu_primal, cov_primal),
            (jnp.zeros_like(eps), mu_tangent, cov_tangent),
        )
        return Dual(primal_out, tangent_out)


mv_normal_reparam = MvNormalREPARAM()


@Pytree.dataclass
class Uniform(TailCallADEVPrimitive):
    def sample(self, key: PRNGKey, *_args):
        v = tfd.Uniform(low=0.0, high=1.0).sample(seed=key)
        return v

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: tuple[Any, ...],
    ):
        key, sub_key = jax.random.split(key)
        x = tfd.Uniform(low=0.0, high=1.0).sample(seed=sub_key)
        return Dual(x, 0.0)


uniform = Uniform()


@Pytree.dataclass
class BetaIMPLICIT(TailCallADEVPrimitive):
    def sample(self, key, *args):
        alpha, beta = args
        v = tfd.Beta(concentration1=alpha, concentration0=beta).sample(seed=key)
        return v

    def before_tail_call(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
    ):
        # Because TFP already overloads their Beta sampler with implicit
        # differentiation rules for JVP, we directly utilize their rules.
        def _inner(alpha, beta):
            # Invoking TFP's Implicit reparametrization:
            # https://github.com/tensorflow/probability/blob/v0.23.0/tensorflow_probability/python/distributions/beta.py#L292-L306
            x = tfd.Beta(concentration1=alpha, concentration0=beta).sample(seed=key)
            return x

        # We invoke JAX's JVP (which utilizes TFP's registered implicit differentiation
        # rule for Beta) to get a primal and tangent out.
        primals = Dual.tree_primal(dual_tree)
        tangents = Dual.tree_tangent(dual_tree)
        primal_out, tangent_out = jax.jvp(_inner, primals, tangents)
        return Dual(primal_out, tangent_out)


beta_implicit = BetaIMPLICIT()


@Pytree.dataclass
class Baseline(ADEVPrimitive):
    prim: ADEVPrimitive

    def sample(self, key, *args):
        return self.prim.sample(key, *args[1:])

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ):
        (kpure, kdual) = konts
        (b_primal, *prim_primals) = Dual.tree_primal(dual_tree)
        (b_tangent, *prim_tangents) = Dual.tree_tangent(dual_tree)

        def new_kdual(key, dual: Dual):
            ret_dual = kdual(key, dual)

            def _inner(ret, b):
                return ret - b

            primal, tangent = jax.jvp(
                _inner,
                (ret_dual.primal, b_primal),
                (ret_dual.tangent, b_tangent),
            )
            return Dual(primal, tangent)

        l_dual = self.prim.jvp_estimate(
            key,
            Dual.dual_tree(prim_primals, prim_tangents),
            (kpure, new_kdual),
        )

        def _inner(left, right):
            return left + right

        primal, tangent = jax.jvp(
            _inner,
            (l_dual.primal, b_primal),
            (l_dual.tangent, b_tangent),
        )
        return Dual(primal, tangent)


def baseline(prim):
    return Baseline(prim)


##################
# Loss primitive #
##################


@Pytree.dataclass
class AddCost(ADEVPrimitive):
    def sample(self, key, *args):
        (w,) = args
        return w

    def jvp_estimate(
        self,
        key: PRNGKey,
        dual_tree: DualTree,
        konts: tuple[Any, ...],
    ) -> Dual:
        (_, kdual) = konts
        (w,) = Dual.tree_primal(dual_tree)
        (w_tangent,) = Dual.tree_tangent(dual_tree)
        l_dual = kdual(key, Dual(None, None))
        return Dual(w + l_dual.primal, w_tangent + l_dual.tangent)


def add_cost(w):
    prim = AddCost()
    prim(w)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from genjax._src.core.generative import (
    GenerativeFunction,
)
from genjax._src.core.typing import (
    Callable,
    TypeVar,
)

R = TypeVar("R")


def RepeatCombinator(
    gen_fn: GenerativeFunction[R], /, *, n: int
) -> GenerativeFunction[R]:
    """
    A combinator that samples from a supplied [`genjax.GenerativeFunction`][] `gen_fn` a fixed number of times, returning a vector of `n` results.

    See [`genjax.repeat`][] for more details.
    """
    return (
        gen_fn.contramap(lambda _idx, args: args)
        .vmap(in_axes=(0, None))
        .contramap(lambda *args: (jnp.zeros(n), args))
    )


def repeat(*, n: int) -> Callable[[GenerativeFunction[R]], GenerativeFunction[R]]:
    """
    Returns a decorator that wraps a [`genjax.GenerativeFunction`][] `gen_fn` of type `a -> b` and returns a new `GenerativeFunction` of type `a -> [b]` that samples from `gen_fn `n` times, returning a vector of `n` results.

    The values traced by each call `gen_fn` will be nested under an integer index that matches the loop iteration index that generated it.

    This combinator is useful for creating multiple samples from the same generative model in a batched manner.

    Args:
        n: The number of times to sample from the generative function.

    Returns:
        A new [`genjax.GenerativeFunction`][] that samples from the original function `n` times.

    Examples:
        ```python exec="yes" html="true" source="material-block" session="repeat"
        import genjax, jax


        @genjax.repeat(n=10)
        @genjax.gen
        def normal_draws(mean):
            return genjax.normal(mean, 1.0) @ "x"


        key = jax.random.key(314159)

        # Generate 10 draws from a normal distribution with mean 2.0
        tr = jax.jit(normal_draws.simulate)(key, (2.0,))
        print(tr.render_html())
        ```
    """

    def decorator(gen_fn: GenerativeFunction[R]) -> GenerativeFunction[R]:
        return RepeatCombinator(gen_fn, n=n)

    return decorator



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import functools
import itertools as it

import jax.core as jc
import jax.tree_util as jtu
from jax import tree_util
from jax import util as jax_util
from jax.extend import linear_util as lu
from jax.interpreters import batching, mlir
from jax.interpreters import partial_eval as pe

from genjax._src.core.interpreters.staging import WrappedFunWithAux, stage
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import Any, Callable

#########################
# Custom JAX primitives #
#########################

# Wrapper to assign a correct type.
batch_subtrace: Callable[[lu.WrappedFun], WrappedFunWithAux] = batching.batch_subtrace  # pyright: ignore[reportAssignmentType]


@lu.transformation
def __batch_fun(in_dims, *in_vals, **params):
    with jc.new_main(batching.BatchTrace, axis_name=jc.no_axis_name) as main:
        out_vals = yield (
            (main, in_dims, *in_vals),
            params,
        )
        del main
    yield out_vals


_batch_fun: Callable[[lu.WrappedFun, Any], lu.WrappedFun] = __batch_fun  # pyright: ignore[reportAssignmentType]


def batch_fun(fun: lu.WrappedFun, in_dims) -> WrappedFunWithAux:
    fun, out_dims = batch_subtrace(fun)
    return _batch_fun(fun, in_dims), out_dims


class FlatPrimitive(jc.Primitive):
    """Contains default implementations of transformations."""

    def __init__(self, name):
        super(FlatPrimitive, self).__init__(name)
        self.multiple_results = True

        def _abstract(*flat_avals, **params):
            return pe.abstract_eval_fun(self.impl, *flat_avals, **params)

        self.def_abstract_eval(_abstract)

        def _batch(args, dims, **params):
            batched, out_dims = batch_fun(lu.wrap_init(self.impl, params), dims)
            return batched.call_wrapped(*args), out_dims()

        batching.primitive_batchers[self] = _batch

        def _mlir(ctx: mlir.LoweringRuleContext, *mlir_args, **params):
            lowering = mlir.lower_fun(self.impl, multiple_results=True)
            return lowering(ctx, *mlir_args, **params)

        mlir.register_lowering(self, _mlir)


class InitialStylePrimitive(FlatPrimitive):
    """Contains default implementations of transformations."""

    def __init__(self, name):
        super().__init__(name)

        def fun_impl(*args, **params):
            consts, args = jax_util.split_list(args, [params["num_consts"]])
            return jc.eval_jaxpr(params["_jaxpr"], consts, *args)

        self.def_impl(fun_impl)


def initial_style_bind(prim, **params):
    """Binds a primitive to a function call."""

    def bind(f):
        """Wraps a function to be bound to a primitive, keeping track of Pytree
        information."""

        def wrapped(*args, **kwargs):
            """Runs a function and binds it to a call primitive."""
            jaxpr, (flat_args, in_tree, out_tree) = stage(f)(*args, **kwargs)
            outs = prim.bind(
                *it.chain(jaxpr.literals, flat_args),
                _jaxpr=jaxpr.jaxpr,
                in_tree=in_tree,
                out_tree=out_tree,
                num_consts=len(jaxpr.literals),
                **params,
            )
            return tree_util.tree_unflatten(out_tree(), outs)

        return wrapped

    return bind


#######################
# Forward interpreter #
#######################

VarOrLiteral = jc.Var | jc.Literal


@Pytree.dataclass
class Environment(Pytree):
    """Keeps track of variables and their values during propagation."""

    env: dict[int, Any] = Pytree.field(default_factory=dict)

    def read(self, var: VarOrLiteral) -> Any:
        if isinstance(var, jc.Literal):
            return var.val
        else:
            v = self.env.get(var.count)
            if v is None:
                raise ValueError(
                    f"Unbound variable in interpreter environment at count {var.count}:\nEnvironment keys (count): {list(self.env.keys())}"
                )
            return v

    def get(self, var: VarOrLiteral) -> Any:
        if isinstance(var, jc.Literal):
            return var.val
        else:
            return self.env.get(var.count)

    def write(self, var: VarOrLiteral, cell: Any) -> Any:
        if isinstance(var, jc.Literal):
            return cell
        cur_cell = self.get(var)
        if isinstance(var, jc.DropVar):
            return cur_cell
        self.env[var.count] = cell
        return self.env[var.count]

    def __getitem__(self, var: VarOrLiteral) -> Any:
        return self.read(var)

    def __setitem__(self, key, val):
        raise ValueError(
            "Environments do not support __setitem__. Please use the "
            "`write` method instead."
        )

    def __contains__(self, var: VarOrLiteral):
        if isinstance(var, jc.Literal):
            return True
        return var.count in self.env

    def copy(self):
        keys = list(self.env.keys())
        return Environment({k: self.env[k] for k in keys})


class StatefulHandler:
    @abc.abstractmethod
    def handles(self, primitive: jc.Primitive) -> bool:
        pass

    @abc.abstractmethod
    def dispatch(
        self,
        primitive: jc.Primitive,
        *args,
        **kwargs,
    ) -> list[Any]:
        pass


@Pytree.dataclass
class ForwardInterpreter(Pytree):
    def _eval_jaxpr_forward(
        self,
        stateful_handler,
        _jaxpr: jc.Jaxpr,
        consts: list[Any],
        args: list[Any],
    ):
        env = Environment()
        jax_util.safe_map(env.write, _jaxpr.constvars, consts)
        jax_util.safe_map(env.write, _jaxpr.invars, args)
        for eqn in _jaxpr.eqns:
            invals = jax_util.safe_map(env.read, eqn.invars)
            subfuns, params = eqn.primitive.get_bind_params(eqn.params)
            args = subfuns + invals
            if stateful_handler.handles(eqn.primitive):
                outvals = stateful_handler.dispatch(eqn.primitive, *args, **params)
            else:
                outvals = eqn.primitive.bind(*args, **params)
            if not eqn.primitive.multiple_results:
                outvals = [outvals]
            jax_util.safe_map(env.write, eqn.outvars, outvals)

        return jax_util.safe_map(env.read, _jaxpr.outvars)

    def run_interpreter(self, stateful_handler, fn, *args, **kwargs):
        def _inner(*args):
            return fn(*args, **kwargs)

        closed_jaxpr, (flat_args, _, out_tree) = stage(_inner)(*args)
        jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.literals
        flat_out = self._eval_jaxpr_forward(
            stateful_handler,
            jaxpr,
            consts,
            flat_args,
        )
        return jtu.tree_unflatten(out_tree(), flat_out)


def forward(f: Callable[..., Any]):
    @functools.wraps(f)
    def wrapped(stateful_handler: StatefulHandler, *args):
        interpreter = ForwardInterpreter()
        return interpreter.run_interpreter(
            stateful_handler,
            f,
            *args,
        )

    return wrapped



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains the `Distribution` abstract base class."""

from abc import abstractmethod

import jax
import jax.numpy as jnp
from jax.experimental import checkify
from tensorflow_probability.substrates import jax as tfp

from genjax._src.checkify import optional_check
from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    Constraint,
    EditRequest,
    EmptyConstraint,
    GenerativeFunction,
    Mask,
    NotSupportedEditRequest,
    Projection,
    R,
    Regenerate,
    Retdiff,
    Score,
    Selection,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.generative.choice_map import ChoiceMapConstraint
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.interpreters.staging import FlagOp, to_shape_fn
from genjax._src.core.pytree import Closure, Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    Generic,
    PRNGKey,
)

tfd = tfp.distributions

#####
# DistributionTrace
#####


@Pytree.dataclass
class DistributionTrace(
    Generic[R],
    Trace[R],
):
    gen_fn: GenerativeFunction[R]
    args: tuple[Any, ...]
    value: R
    score: Score

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_retval(self) -> R:
        return self.value

    def get_gen_fn(self) -> GenerativeFunction[R]:
        return self.gen_fn

    def get_score(self) -> Score:
        return self.score

    def get_choices(self) -> ChoiceMap:
        return ChoiceMap.choice(self.value)


################
# Distribution #
################


class Distribution(Generic[R], GenerativeFunction[R]):
    @abstractmethod
    def random_weighted(
        self,
        key: PRNGKey,
        *args,
    ) -> tuple[Score, R]:
        pass

    @abstractmethod
    def estimate_logpdf(
        self,
        key: PRNGKey,
        v: R,
        *args,
    ) -> Score:
        pass

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> Trace[R]:
        (w, v) = self.random_weighted(key, *args)
        tr = DistributionTrace(self, args, v, w)
        return tr

    def generate_choice_map(
        self,
        key: PRNGKey,
        chm: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Trace[R], Weight]:
        v = chm.get_value()
        match v:
            case None:
                tr = self.simulate(key, args)
                return tr, jnp.array(0.0)

            case Mask(value, flag):

                def _simulate(key, v):
                    score, new_v = self.random_weighted(key, *args)
                    w = 0.0
                    return (score, w, new_v)

                def _importance(key, v):
                    w = self.estimate_logpdf(key, v, *args)
                    return (w, w, v)

                score, w, new_v = jax.lax.cond(flag, _importance, _simulate, key, value)
                tr = DistributionTrace(self, args, new_v, score)
                return tr, w

            case _:
                w = self.estimate_logpdf(key, v, *args)
                tr = DistributionTrace(self, args, v, w)
                return tr, w

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[Trace[R], Weight]:
        match constraint:
            case ChoiceMapConstraint(chm):
                tr, w = self.generate_choice_map(key, chm, args)
            case EmptyConstraint():
                tr = self.simulate(key, args)
                w = jnp.array(0.0)
            case _:
                raise Exception("Unhandled type.")
        return tr, w

    def edit_empty(
        self,
        trace: Trace[R],
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], Update]:
        sample = trace.get_choices()
        primals = Diff.tree_primal(argdiffs)
        new_score, _ = self.assess(sample, primals)
        new_trace = DistributionTrace(self, primals, sample.get_value(), new_score)
        return (
            new_trace,
            new_score - trace.get_score(),
            Diff.no_change(trace.get_retval()),
            Update(ChoiceMap.empty()),
        )

    def edit_update_with_constraint(
        self,
        key: PRNGKey,
        trace: Trace[R],
        constraint: Constraint,
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], Update]:
        primals = Diff.tree_primal(argdiffs)
        match constraint:
            case ChoiceMapConstraint(chm):
                match chm.get_value():
                    case Mask() as masked_value:

                        def _true_branch(key, new_value: R, _):
                            fwd = self.estimate_logpdf(key, new_value, *primals)
                            bwd = trace.get_score()
                            w = fwd - bwd
                            return (new_value, w, fwd)

                        def _false_branch(key, _, old_value: R):
                            fwd = self.estimate_logpdf(key, old_value, *primals)
                            bwd = trace.get_score()
                            w = fwd - bwd
                            return (old_value, w, fwd)

                        flag = masked_value.primal_flag()
                        new_value: R = masked_value.value
                        old_choices = trace.get_choices()
                        old_value: R = old_choices.get_value()

                        new_value, w, score = FlagOp.cond(
                            flag,
                            _true_branch,
                            _false_branch,
                            key,
                            new_value,
                            old_value,
                        )
                        return (
                            DistributionTrace(self, primals, new_value, score),
                            w,
                            Diff.unknown_change(new_value),
                            Update(
                                old_choices.mask(flag),
                            ),
                        )
                    case None:
                        value_chm = trace.get_choices()
                        v = value_chm.get_value()
                        fwd = self.estimate_logpdf(key, v, *primals)
                        bwd = trace.get_score()
                        w = fwd - bwd
                        new_tr = DistributionTrace(self, primals, v, fwd)
                        retval_diff = Diff.no_change(v)
                        return (new_tr, w, retval_diff, Update(ChoiceMap.empty()))

                    case v:
                        fwd = self.estimate_logpdf(key, v, *primals)
                        bwd = trace.get_score()
                        w = fwd - bwd
                        new_tr = DistributionTrace(self, primals, v, fwd)
                        discard = trace.get_choices()
                        retval_diff = Diff.unknown_change(v)
                        return (new_tr, w, retval_diff, Update(discard))
            case _:
                raise Exception(f"Unhandled constraint in edit: {type(constraint)}.")

    def project(
        self,
        key: PRNGKey,
        trace: Trace[R],
        projection: Projection[Any],
    ) -> Weight:
        assert isinstance(projection, Selection)
        return jnp.where(
            projection.check(),
            trace.get_score(),
            jnp.array(0.0),
        )

    def edit_regenerate(
        self,
        key: PRNGKey,
        trace: Trace[R],
        selection: Selection,
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], EditRequest]:
        check = () in selection
        if FlagOp.concrete_true(check):
            primals = Diff.tree_primal(argdiffs)
            w, new_v = self.random_weighted(key, *primals)
            incremental_w = w - trace.get_score()
            old_v = trace.get_retval()
            new_trace = DistributionTrace(self, primals, new_v, w)
            return (
                new_trace,
                incremental_w,
                Diff.unknown_change(new_v),
                Update(ChoiceMap.choice(old_v)),
            )
        elif FlagOp.concrete_false(check):
            if Diff.static_check_no_change(argdiffs):
                return (
                    trace,
                    jnp.array(0.0),
                    Diff.no_change(trace.get_retval()),
                    Update(ChoiceMap.empty()),
                )
            else:
                chm = trace.get_choices()
                primals = Diff.tree_primal(argdiffs)
                new_score, _ = self.assess(chm, primals)
                new_trace = DistributionTrace(self, primals, chm.get_value(), new_score)
                return (
                    new_trace,
                    new_score - trace.get_score(),
                    Diff.no_change(trace.get_retval()),
                    Update(
                        ChoiceMap.empty(),
                    ),
                )
        else:
            raise NotImplementedError

    def edit_update(
        self,
        key: PRNGKey,
        trace: Trace[R],
        constraint: Constraint,
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], Update]:
        match constraint:
            case EmptyConstraint():
                return self.edit_empty(trace, argdiffs)

            case ChoiceMapConstraint():
                return self.edit_update_with_constraint(
                    key, trace, constraint, argdiffs
                )

            case _:
                raise Exception(f"Not implement fwd problem: {constraint}.")

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[R],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[Trace[R], Weight, Retdiff[R], EditRequest]:
        match edit_request:
            case Update(chm):
                return self.edit_update(
                    key,
                    trace,
                    ChoiceMapConstraint(chm),
                    argdiffs,
                )
            case Regenerate(selection):
                return self.edit_regenerate(
                    key,
                    trace,
                    selection,
                    argdiffs,
                )

            case _:
                raise NotSupportedEditRequest(edit_request)

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ):
        raise NotImplementedError


################
# ExactDensity #
################

_fake_key = jnp.array([0, 0], dtype=jnp.uint32)


class ExactDensity(Generic[R], Distribution[R]):
    @abstractmethod
    def sample(self, key: PRNGKey, *args) -> R:
        pass

    @abstractmethod
    def logpdf(self, v: R, *args) -> Score:
        pass

    def __abstract_call__(self, *args):
        return to_shape_fn(self.sample, jnp.zeros)(_fake_key, *args)

    def handle_kwargs(self) -> GenerativeFunction[R]:
        @Pytree.partial(self)
        def sample_with_kwargs(self: "ExactDensity[R]", key, args, kwargs):
            return self.sample(key, *args, **kwargs)

        @Pytree.partial(self)
        def logpdf_with_kwargs(self: "ExactDensity[R]", v, args, kwargs):
            return self.logpdf(v, *args, **kwargs)

        return ExactDensityFromCallables(
            sample_with_kwargs,
            logpdf_with_kwargs,
        )

    def random_weighted(
        self,
        key: PRNGKey,
        *args,
    ) -> tuple[Score, R]:
        """
        Given arguments to the distribution, sample from the distribution, and return the exact log density of the sample, and the sample.
        """
        v = self.sample(key, *args)
        w = self.estimate_logpdf(key, v, *args)
        return (w, v)

    def estimate_logpdf(
        self,
        key: PRNGKey,
        v: R,
        *args,
    ) -> Weight:
        """
        Given a sample and arguments to the distribution, return the exact log density of the sample.
        """
        w = self.logpdf(v, *args)
        if w.shape:
            return jnp.sum(w)
        else:
            return w

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Weight, R]:
        key = jax.random.key(0)
        v = sample.get_value()
        match v:
            case Mask(value, flag):

                def _check():
                    checkify.check(
                        bool(flag),
                        "Attempted to unmask when a mask flag is False: the masked value is invalid.\n",
                    )

                optional_check(_check)
                w = self.estimate_logpdf(key, value, *args)
                return w, value
            case _:
                w = self.estimate_logpdf(key, v, *args)
                return w, v


@Pytree.dataclass
class ExactDensityFromCallables(Generic[R], ExactDensity[R]):
    sampler: Closure[R]
    logpdf_evaluator: Closure[Score]

    def sample(self, key, *args) -> R:
        return self.sampler(key, *args)

    def logpdf(self, v, *args) -> Score:
        return self.logpdf_evaluator(v, *args)


def exact_density(
    sample: Closure[R] | Callable[..., R],
    logpdf: Closure[Score] | Callable[..., Score],
):
    if not isinstance(sample, Closure):
        sample = Closure[R]((), sample)

    if not isinstance(logpdf, Closure):
        logpdf = Closure[Score]((), logpdf)

    return ExactDensityFromCallables[R](sample, logpdf)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from genjax._src.core.generative import (
    Address,
    AddressComponent,
    Argdiffs,
    Arguments,
    ChoiceMap,
    ChoiceMapBuilder,
    Constraint,
    EditRequest,
    EmptyConstraint,
    GenerativeFunction,
    GenerativeFunctionClosure,
    Mask,
    MaskedConstraint,
    R,
    Retdiff,
    Score,
    Selection,
    SelectionBuilder,
    Trace,
    Update,
    Weight,
)

__all__ = [
    "Address",
    "AddressComponent",
    "Argdiffs",
    "Arguments",
    "ChoiceMap",
    "ChoiceMapBuilder",
    "Constraint",
    "EditRequest",
    "EmptyConstraint",
    "GenerativeFunction",
    "GenerativeFunctionClosure",
    "Mask",
    "MaskedConstraint",
    "R",
    "Retdiff",
    "Score",
    "Selection",
    "SelectionBuilder",
    "Trace",
    "Update",
    "Weight",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    Constraint,
    EditRequest,
    GenerativeFunction,
    Projection,
    Retdiff,
    Score,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.interpreters.incremental import Diff, NoChange, UnknownChange
from genjax._src.core.interpreters.staging import multi_switch, tree_choose
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    FloatArray,
    Generic,
    IntArray,
    PRNGKey,
    TypeVar,
)

R = TypeVar("R")

################
# Switch trace #
################


@Pytree.dataclass
class SwitchTrace(Generic[R], Trace[R]):
    gen_fn: "SwitchCombinator[R]"
    args: tuple[Any, ...]
    subtraces: list[Trace[R]]
    retval: R
    score: FloatArray

    def get_idx(self) -> int | IntArray:
        """
        Get the index used to select the branch in this SwitchTrace.

        Returns:
            The index value used to select the executed branch.

        Note:
            This method assumes that the first argument passed to the SwitchCombinator was the index used for branch selection.
        """
        return self.get_args()[0]

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_choices(self) -> ChoiceMap:
        idx = self.get_idx()
        sub_chms = (tr.get_choices() for tr in self.subtraces)
        return ChoiceMap.switch(idx, sub_chms)

    def get_gen_fn(self):
        return self.gen_fn

    def get_retval(self):
        return self.retval

    def get_score(self):
        return self.score


#####################
# Switch combinator #
#####################


@Pytree.dataclass
class SwitchCombinator(Generic[R], GenerativeFunction[R]):
    """
    `SwitchCombinator` accepts `n` generative functions as input and returns a new [`genjax.GenerativeFunction`][] that accepts `n+1` arguments:

    - an index in the range `[0, n-1]`
    - a tuple of arguments for each of the input generative functions

    and executes the generative function at the supplied index with its provided arguments.

    If `index` is out of bounds, `index` is clamped to within bounds.

    !!! info "Existence uncertainty"

        This pattern allows `GenJAX` to express existence uncertainty over random choices -- as different generative function branches need not share addresses.

    Attributes:
        branches: generative functions that the `SwitchCombinator` will select from based on the supplied index.

    Examples:
        Create a `SwitchCombinator` via the [`genjax.switch`][] method:
        ```python exec="yes" html="true" source="material-block" session="switch"
        import jax, genjax


        @genjax.gen
        def branch_1():
            x = genjax.normal(0.0, 1.0) @ "x1"


        @genjax.gen
        def branch_2():
            x = genjax.bernoulli(0.3) @ "x2"


        switch = genjax.switch(branch_1, branch_2)

        key = jax.random.key(314159)
        jitted = jax.jit(switch.simulate)

        # Select `branch_2` by providing 1:
        tr = jitted(key, (1, (), ()))

        print(tr.render_html())
        ```
    """

    branches: tuple[GenerativeFunction[R], ...]

    def _indices(self):
        return range(len(self.branches))

    def __abstract_call__(self, *args) -> R:
        idx, args = args[0], args[1:]
        retvals = list(
            f.__abstract_call__(*f_args) for f, f_args in zip(self.branches, args)
        )
        return tree_choose(idx, retvals)

    def _check_args_match_branches(self, args):
        assert len(args) == len(self.branches)

    ## Simulate methods

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> SwitchTrace[R]:
        idx, branch_args = args[0], args[1:]
        self._check_args_match_branches(branch_args)

        fs = list(f.simulate for f in self.branches)
        f_args = list((key, args) for args in branch_args)

        subtraces = multi_switch(idx, fs, f_args)
        retval, score = tree_choose(
            idx, list((tr.get_retval(), tr.get_score()) for tr in subtraces)
        )
        return SwitchTrace(self, args, subtraces, retval, score)

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Score, R]:
        idx, branch_args = args[0], args[1:]
        self._check_args_match_branches(branch_args)

        fs = list(f.assess for f in self.branches)
        f_args = list((sample, args) for args in branch_args)

        return tree_choose(idx, multi_switch(idx, fs, f_args))

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[SwitchTrace[R], Weight]:
        idx, branch_args = args[0], args[1:]
        self._check_args_match_branches(branch_args)

        fs = list(f.generate for f in self.branches)
        f_args = list((key, constraint, args) for args in branch_args)

        pairs = multi_switch(idx, fs, f_args)
        subtraces = list(tr for tr, _ in pairs)

        retval, score, weight = tree_choose(
            idx, list((tr.get_retval(), tr.get_score(), w) for tr, w in pairs)
        )
        return SwitchTrace(self, args, subtraces, retval, score), weight

    def project(
        self,
        key: PRNGKey,
        trace: Trace[R],
        projection: Projection[Any],
    ) -> Weight:
        assert isinstance(trace, SwitchTrace)
        idx = trace.get_idx()

        fs = list(f.project for f in self.branches)
        f_args = list((key, tr, projection) for tr in trace.subtraces)

        return tree_choose(idx, multi_switch(idx, fs, f_args))

    def _make_edit_fresh_trace(self, gen_fn: GenerativeFunction[R]):
        """
        Creates a function to handle editing a fresh trace when the switch index changes.

        This method is used internally by the `edit` method to handle cases where
        the switch index has changed, requiring the generation of a new trace
        for the selected branch.
        """

        def inner(
            key: PRNGKey,
            edit_request: Update,
            argdiffs: Argdiffs,
        ) -> tuple[Trace[R], Weight, Retdiff[R], EditRequest]:
            # the old trace only has a filled-in subtrace for the original index. All other subtraces are filled with zeros. In the case of a changed index we need to
            #
            # - generate a fresh trace for the new branch,
            # - call `edit` with that new trace (setting the argdiffs passed into `edit` as `no_change`, since we used the same args to create the new trace)
            # - return the edit result with the `retdiff` wrapped in `unknown_change` (since our return value comes from a new branch)
            primals = Diff.tree_primal(argdiffs)
            new_trace = gen_fn.simulate(key, primals)

            tr, w, rd, bwd_request = gen_fn.edit(
                key,
                new_trace,
                edit_request,
                Diff.no_change(argdiffs),
            )
            return tr, w, Diff.unknown_change(rd), bwd_request

        return inner

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[R],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[SwitchTrace[R], Weight, Retdiff[R], EditRequest]:
        assert isinstance(edit_request, Update)
        assert isinstance(trace, SwitchTrace)

        idx_diff, branch_argdiffs = argdiffs[0], argdiffs[1:]
        self._check_args_match_branches(branch_argdiffs)

        primals = Diff.tree_primal(argdiffs)
        new_idx = primals[0]

        if Diff.tree_tangent(idx_diff) == NoChange:
            # If the index hasn't changed, perform edits on each branch.
            fs = list(f.edit for f in self.branches)
            f_args = list(
                (key, trace, edit_request, argdiffs)
                for trace, argdiffs in zip(trace.subtraces, branch_argdiffs)
            )
        else:
            fs = list(self._make_edit_fresh_trace(f) for f in self.branches)
            f_args = list((key, edit_request, argdiffs) for argdiffs in branch_argdiffs)

        rets = multi_switch(new_idx, fs, f_args)

        subtraces = list(t[0] for t in rets)
        score, weight, retdiff = tree_choose(
            new_idx, list((tr.get_score(), w, rd) for tr, w, rd, _ in rets)
        )
        retval: R = Diff.tree_primal(retdiff)

        if Diff.tree_tangent(idx_diff) == UnknownChange:
            weight += score - trace.get_score()

        # TODO: this is totally wrong, fix in future PR.
        bwd_request: Update = rets[0][3]

        return (
            SwitchTrace(self, primals, subtraces, retval, score),
            weight,
            retdiff,
            bwd_request,
        )


#############
# Decorator #
#############


def switch(
    *gen_fns: GenerativeFunction[R],
) -> SwitchCombinator[R]:
    """
    Given `n` [`genjax.GenerativeFunction`][] inputs, returns a [`genjax.GenerativeFunction`][] that accepts `n+1` arguments:

    - an index in the range $[0, n)$
    - a tuple of arguments for each of the input generative functions (`n` total tuples)

    and executes the generative function at the supplied index with its provided arguments.

    If `index` is out of bounds, `index` is clamped to within bounds.

    Args:
        gen_fns: generative functions that the `SwitchCombinator` will select from.

    Examples:
        Create a `SwitchCombinator` via the [`genjax.switch`][] method:
        ```python exec="yes" html="true" source="material-block" session="switch"
        import jax, genjax


        @genjax.gen
        def branch_1():
            x = genjax.normal(0.0, 1.0) @ "x1"


        @genjax.gen
        def branch_2():
            x = genjax.bernoulli(0.3) @ "x2"


        switch = genjax.switch(branch_1, branch_2)

        key = jax.random.key(314159)
        jitted = jax.jit(switch.simulate)

        # Select `branch_2` by providing 1:
        tr = jitted(key, (1, (), ()))

        print(tr.render_html())
        ```
    """
    return SwitchCombinator[R](gen_fns)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains a set of types and type aliases which are used throughout the
codebase.

Type annotations in the codebase are exported out of this module for consistency.
"""

import sys
from types import EllipsisType
from typing import Annotated

import beartype.typing as btyping
import jax.numpy as jnp
import jaxtyping as jtyping
import numpy as np
from beartype.vale import Is
from jax import core as jc

if sys.version_info >= (3, 11, 0):
    from typing import Self
else:
    from typing_extensions import Self

Any = btyping.Any
PRNGKey = jtyping.PRNGKeyArray
Array = jtyping.Array
ArrayLike = jtyping.ArrayLike
IntArray = jtyping.Int[jtyping.Array, "..."]
FloatArray = jtyping.Float[jtyping.Array, "..."]
BoolArray = jtyping.Bool[jtyping.Array, "..."]
Callable = btyping.Callable
TypeAlias = btyping.TypeAlias
Sequence = btyping.Sequence
Iterable = btyping.Iterable
Final = btyping.Final
Generator = btyping.Generator
Literal = btyping.Literal

# JAX Type alias.
InAxes = int | Sequence[Any] | None

Flag = bool | BoolArray

#################################
# Trace-time-checked primitives #
#################################

ScalarShaped = Is[lambda arr: jnp.array(arr, copy=False).shape == ()]
ScalarFlag = Annotated[Flag, ScalarShaped]
ScalarInt = Annotated[IntArray, ScalarShaped]

############
# Generics #
############

Generic = btyping.Generic
TypeVar = btyping.TypeVar
ParamSpec = btyping.ParamSpec


#################
# Static checks #
#################


def static_check_is_array(v: Any) -> bool:
    return (
        isinstance(v, jnp.ndarray)
        or isinstance(v, np.ndarray)
        or isinstance(v, jc.Tracer)
    )


def static_check_is_concrete(x: Any) -> bool:
    return not isinstance(x, jc.Tracer)


# TODO: the dtype comparison needs to be replaced with something
# more robust.
def static_check_supports_grad(v):
    return static_check_is_array(v) and v.dtype == np.float32


def static_check_shape_dtype_equivalence(vs: list[Array]) -> bool:
    shape_dtypes = [(v.shape, v.dtype) for v in vs]
    num_unique = set(shape_dtypes)
    return len(num_unique) == 1


__all__ = [
    "Annotated",
    "Any",
    "Array",
    "ArrayLike",
    "BoolArray",
    "Callable",
    "EllipsisType",
    "Final",
    "Flag",
    "FloatArray",
    "Generator",
    "Generic",
    "InAxes",
    "IntArray",
    "Is",
    "Iterable",
    "PRNGKey",
    "ParamSpec",
    "ScalarFlag",
    "ScalarInt",
    "ScalarShaped",
    "Self",
    "Sequence",
    "TypeAlias",
    "TypeVar",
    "static_check_is_array",
    "static_check_is_concrete",
    "static_check_shape_dtype_equivalence",
    "static_check_supports_grad",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import treescope


def pretty():
    treescope.register_as_default()
    treescope.register_autovisualize_magic()

    # Optional: enables automatic array visualization
    treescope.active_autovisualizer.set_interactive(treescope.ArrayAutovisualizer())


__all__ = [
    "pretty",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

import jax

from genjax._src.core.generative import (
    ChoiceMap,
    GenerativeFunction,
    Selection,
    Weight,
)
from genjax._src.core.generative.core import Score
from genjax._src.core.generative.generative_function import Trace
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Annotated,
    Any,
    Callable,
    Generic,
    Is,
    PRNGKey,
    TypeVar,
)
from genjax._src.generative_functions.distributions.distribution import Distribution

R = TypeVar("R")

####################
# Posterior target #
####################


def validate_non_marginal(x):
    if isinstance(x, Marginal):
        raise TypeError("Target does not support Marginal generative functions.")
    return True


@Pytree.dataclass
class Target(Generic[R], Pytree):
    """
    A `Target` represents an unnormalized target distribution induced by conditioning a generative function on a [`genjax.Constraint`][].

    Targets are created by providing a generative function, arguments to the generative function, and a constraint.

    Examples:
        Creating a target from a generative function, by providing arguments and a constraint:
        ```python exec="yes" html="true" source="material-block" session="core"
        import genjax
        from genjax import ChoiceMapBuilder as C
        from genjax.inference import Target


        @genjax.gen
        def model():
            x = genjax.normal(0.0, 1.0) @ "x"
            y = genjax.normal(x, 1.0) @ "y"
            return x


        target = Target(model, (), C["y"].set(3.0))
        print(target.render_html())
        ```
    """

    p: Annotated[GenerativeFunction[R], Is[validate_non_marginal]]
    args: tuple[Any, ...]
    constraint: ChoiceMap

    def importance(
        self, key: PRNGKey, constraint: ChoiceMap
    ) -> tuple[Trace[R], Weight]:
        merged = self.constraint.merge(constraint)
        return self.p.importance(key, merged, self.args)

    def filter_to_unconstrained(self, choice_map):
        selection = ~self.constraint.get_selection()
        return choice_map.filter(selection)

    def __getitem__(self, addr):
        return self.constraint[addr]


#######################
# Sample distribution #
#######################

SampleDistribution = Distribution[ChoiceMap]
"""
The abstract class `SampleDistribution` represents the type of distributions whose return value type is a `ChoiceMap`. This is the abstract base class of `Algorithm`, as well as `Marginal`.
"""

########################
# Inference algorithms #
########################


class Algorithm(Generic[R], SampleDistribution):
    """`Algorithm` is the type of inference
    algorithms: probabilistic programs which provide interfaces for sampling from
    posterior approximations, and estimating densities.

    **The stochastic probability interface for `Algorithm`**

    Inference algorithms implement the stochastic probability interface:

    * `Algorithm.random_weighted` exposes sampling from the approximation
    which the algorithm represents: it accepts a `Target` as input, representing the
    unnormalized distribution, and returns a sample from an approximation to
    the normalized distribution, along with a density estimate of the normalized distribution.

    * `Algorithm.estimate_logpdf` exposes density estimation for the
    approximation which `Algorithm.random_weighted` samples from:
    it accepts a value on the support of the approximation, and the `Target` which
    induced the approximation as input, and returns an estimate of the density of
    the approximation.

    **Optional methods for gradient estimators**

    Subclasses of type `Algorithm` can also implement two optional methods
    designed to support effective gradient estimators for variational objectives
    (`estimate_normalizing_constant` and `estimate_reciprocal_normalizing_constant`).
    """

    #########
    # GenSP #
    #########

    @abstractmethod
    def random_weighted(
        self,
        key: PRNGKey,
        *args: Any,
    ) -> tuple[Score, ChoiceMap]:
        """
        Given a [`Target`][genjax.inference.Target], return a [`ChoiceMap`][genjax.core.ChoiceMap] from an approximation to the normalized distribution of the target, and a random [`Weight`][genjax.core.Weight] estimate of the normalized density of the target at the sample.

        The `sample` is a sample on the support of `target.gen_fn` which _are not in_ `target.constraints`, produced by running the inference algorithm.

        Let $T_P(a, c)$ denote the target, with $P$ the distribution on samples represented by `target.gen_fn`, and $S$ denote the sample. Let $w$ denote the weight `w`. The weight $w$ is a random weight such that $w$ satisfies:

        $$
        \\mathbb{E}\\big[\\frac{1}{w} \\mid S \\big] = \\frac{1}{P(S \\mid c; a)}
        $$

        This interface corresponds to **(Defn 3.2) Unbiased Density Sampler** in [[Lew23](https://dl.acm.org/doi/pdf/10.1145/3591290)].
        """
        assert isinstance(args[0], Target)

    @abstractmethod
    def estimate_logpdf(
        self, key: PRNGKey, v: ChoiceMap, *args: tuple[Any, ...]
    ) -> Score:
        """
        Given a [`ChoiceMap`][genjax.core.ChoiceMap] and a [`Target`][genjax.inference.Target], return a random [`Weight`][genjax.core.Weight] estimate of the normalized density of the target at the sample.

        Let $T_P(a, c)$ denote the target, with $P$ the distribution on samples represented by `target.gen_fn`, and $S$ denote the sample. Let $w$ denote the weight `w`. The weight $w$ is a random weight such that $w$ satisfies:

        $$
        \\mathbb{E}[w] = P(S \\mid c, a)
        $$

        This interface corresponds to **(Defn 3.1) Positive Unbiased Density Estimator** in [[Lew23](https://dl.acm.org/doi/pdf/10.1145/3591290)].
        """

    ################
    # VI via GRASP #
    ################

    @abstractmethod
    def estimate_normalizing_constant(
        self,
        key: PRNGKey,
        target: Target[R],
    ) -> Weight:
        pass

    @abstractmethod
    def estimate_reciprocal_normalizing_constant(
        self,
        key: PRNGKey,
        target: Target[R],
        latent_choices: ChoiceMap,
        w: Weight,
    ) -> Weight:
        pass


############
# Marginal #
############


@Pytree.dataclass
class Marginal(Generic[R], SampleDistribution):
    """The `Marginal` class represents the marginal distribution of a generative function over
    a selection of addresses.
    """

    gen_fn: GenerativeFunction[R]
    selection: Selection = Pytree.field(default=Selection.all())
    algorithm: Algorithm[R] | None = Pytree.field(default=None)

    def random_weighted(
        self,
        key: PRNGKey,
        *args: Any,
    ) -> tuple[Score, ChoiceMap]:
        key, sub_key = jax.random.split(key)
        tr = self.gen_fn.simulate(sub_key, args)
        choices: ChoiceMap = tr.get_choices()
        latent_choices = choices.filter(self.selection)
        key, sub_key = jax.random.split(key)
        bwd_request = ~self.selection
        weight = tr.project(sub_key, bwd_request)
        if self.algorithm is None:
            return weight, latent_choices
        else:
            target = Target(self.gen_fn, args, latent_choices)
            other_choices = choices.filter(~self.selection)
            Z = self.algorithm.estimate_reciprocal_normalizing_constant(
                key, target, other_choices, weight
            )

            return (Z, latent_choices)

    def estimate_logpdf(
        self,
        key: PRNGKey,
        v: ChoiceMap,
        *args: tuple[Any, ...],
    ) -> Score:
        if self.algorithm is None:
            _, weight = self.gen_fn.importance(key, v, args)
            return weight
        else:
            target = Target(self.gen_fn, args, v)
            Z = self.algorithm.estimate_normalizing_constant(key, target)
            return Z


################################
# Inference construct language #
################################


def marginal(
    selection: Selection = Selection.all(),
    algorithm: Algorithm[R] | None = None,
) -> Callable[[GenerativeFunction[R]], Marginal[R]]:
    def decorator(
        gen_fn: GenerativeFunction[R],
    ) -> Marginal[R]:
        return Marginal(
            gen_fn,
            selection,
            algorithm,
        )

    return decorator



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from abc import abstractmethod
from dataclasses import dataclass

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    ChoiceMapConstraint,
    Constraint,
    EditRequest,
    EmptyRequest,
    GenerativeFunction,
    NotSupportedEditRequest,
    PrimitiveEditRequest,
    Projection,
    Regenerate,
    Retdiff,
    Score,
    Selection,
    StaticAddress,
    StaticAddressComponent,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.generative.generative_function import R, push_trace_overload_stack
from genjax._src.core.interpreters.forward import (
    InitialStylePrimitive,
    StatefulHandler,
    forward,
    initial_style_bind,
)
from genjax._src.core.interpreters.incremental import (
    Diff,
    incremental,
)
from genjax._src.core.interpreters.staging import to_shape_fn
from genjax._src.core.pytree import Closure, Const, Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    FloatArray,
    Generic,
    PRNGKey,
    TypeAlias,
)

_WRAPPER_ASSIGNMENTS = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)


# Usage in transforms: checks for duplicate addresses.
@Pytree.dataclass
class AddressVisitor(Pytree):
    visited: list[StaticAddress] = Pytree.static(default_factory=list)

    def visit(self, addr: StaticAddress):
        if addr in self.visited:
            raise AddressReuse(addr)
        else:
            self.visited.append(addr)

    def get_visited(self):
        return self.visited


def collapse_address(
    addr: StaticAddressComponent | StaticAddress,
) -> StaticAddressComponent | StaticAddress:
    return addr[0] if isinstance(addr, tuple) and len(addr) == 1 else addr


#########
# Trace #
#########


@Pytree.dataclass
class StaticTrace(Generic[R], Trace[R]):
    gen_fn: "StaticGenerativeFunction[R]"
    args: tuple[Any, ...]
    retval: R
    addresses: AddressVisitor
    subtraces: list[Trace[Any]]

    def get_args(self) -> tuple[Any, ...]:
        return self.args

    def get_retval(self) -> R:
        return self.retval

    def get_gen_fn(self) -> GenerativeFunction[R]:
        return self.gen_fn

    def get_choices(self) -> ChoiceMap:
        addresses = self.addresses.get_visited()
        sub_chms = (tr.get_choices() for tr in self.subtraces)
        return ChoiceMap.from_mapping(zip(addresses, sub_chms))

    def get_score(self) -> Score:
        return jnp.sum(
            jnp.array([tr.get_score() for tr in self.subtraces], copy=False),
        )

    def get_subtrace(self, addr: StaticAddress):
        addresses = self.addresses.get_visited()
        idx = addresses.index(addr)
        return self.subtraces[idx]


####################################
# Static (trie-like) edit request  #
####################################

StaticDict: TypeAlias = dict[StaticAddressComponent | StaticAddress, EditRequest]


@Pytree.dataclass(match_args=True)
class StaticRequest(PrimitiveEditRequest):
    addressed: StaticDict


##############################
# Static language exceptions #
##############################


class AddressReuse(Exception):
    """Attempt to re-write an address in a GenJAX trace.

    Any given address for a random choice may only be written to once. You can choose a
    different name for the choice, or nest it into a scope where it is unique.
    """


##############
# Primitives #
##############

# Generative function trace intrinsic.
trace_p = InitialStylePrimitive("trace")


############################################################
# Trace call (denotes invocation of a generative function) #
############################################################


# We defer the abstract call here so that, when we
# stage, any traced values stored in `gen_fn`
# get lifted to by `get_shaped_aval`.
def _abstract_gen_fn_call(
    _: tuple[Const[StaticAddress], ...],
    gen_fn: GenerativeFunction[R],
    args: tuple[Any, ...],
):
    return gen_fn.__abstract_call__(*args)


def trace(
    addr: StaticAddress,
    gen_fn: GenerativeFunction[R],
    args: tuple[Any, ...],
):
    """Invoke a generative function, binding its generative semantics with the
    current caller.

    Arguments:
        addr: An address denoting the site of a generative function invocation.
        gen_fn: A generative function invoked as a callee of `StaticGenerativeFunction`.

    """
    addr = Pytree.tree_const(addr)
    return initial_style_bind(trace_p)(_abstract_gen_fn_call)(
        addr,
        gen_fn,
        args,
    )


######################################
#  Generative function interpreters  #
######################################


###########################
# Static language handler #
###########################


# This explicitly makes assumptions about some common fields:
# e.g. it assumes if you are using `StaticHandler.get_submap`
# in your code, that your derived instance has a `constraints` field.
@dataclass
class StaticHandler(StatefulHandler):
    @abstractmethod
    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[R],
        args: tuple[Any, ...],
    ):
        pass

    def handle_retval(self, v):
        return jtu.tree_leaves(v)

    # By default, the interpreter handlers for this language
    # handle the two primitives we defined above
    # (`trace_p`, for random choices)
    def handles(self, primitive):
        return primitive == trace_p

    def dispatch(self, primitive, *tracers, **_params):
        in_tree = _params["in_tree"]
        num_consts = _params.get("num_consts", 0)
        non_const_tracers = tracers[num_consts:]
        addr, gen_fn, args = jtu.tree_unflatten(in_tree, non_const_tracers)
        addr = Pytree.tree_const_unwrap(addr)
        if primitive == trace_p:
            v = self.handle_trace(addr, gen_fn, args)
            return self.handle_retval(v)
        else:
            raise Exception("Illegal primitive: {}".format(primitive))


############
# Simulate #
############


@dataclass
class SimulateHandler(StaticHandler):
    key: PRNGKey
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)
    address_traces: list[Trace[Any]] = Pytree.field(default_factory=list)
    key_counter: int = Pytree.static(default=1)

    def fresh_key_and_increment(self):
        new_key = jax.random.fold_in(self.key, self.key_counter)
        self.key_counter += 1
        return new_key

    def visit(self, addr):
        self.address_visitor.visit(addr)

    def yield_state(self):
        return (
            self.address_visitor,
            self.address_traces,
        )

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        self.visit(addr)
        sub_key = self.fresh_key_and_increment()
        tr = gen_fn.simulate(sub_key, args)
        self.address_traces.append(tr)
        v = tr.get_retval()
        return v


def simulate_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(key, args):
        stateful_handler = SimulateHandler(key)
        retval = forward(source_fn)(stateful_handler, *args)
        (
            address_visitor,
            address_traces,
        ) = stateful_handler.yield_state()
        return (
            args,
            retval,
            address_visitor,
            address_traces,
        )

    return wrapper


##########
# Assess #
##########


@dataclass
class AssessHandler(StaticHandler):
    choice_map_sample: ChoiceMap
    score: Score = Pytree.field(default_factory=lambda: jnp.zeros(()))
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)

    def yield_state(self):
        return (self.score,)

    def get_subsample(self, addr: StaticAddress) -> ChoiceMap:
        return self.choice_map_sample(addr)

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        submap = self.get_subsample(addr)
        (score, v) = gen_fn.assess(submap, args)
        self.score += score
        return v


def assess_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(choice_map_sample: ChoiceMap, args):
        stateful_handler = AssessHandler(choice_map_sample)
        retval = forward(source_fn)(stateful_handler, *args)
        (score,) = stateful_handler.yield_state()
        return (retval, score)

    return wrapper


############################
# Generate request handler #
############################


@dataclass
class GenerateHandler(StaticHandler):
    key: PRNGKey
    choice_map: ChoiceMap
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)
    weight: Weight = Pytree.field(default_factory=lambda: jnp.zeros(()))
    address_traces: list[Trace[Any]] = Pytree.field(default_factory=list)
    key_counter: int = Pytree.static(default=1)

    def fresh_key_and_increment(self):
        new_key = jax.random.fold_in(self.key, self.key_counter)
        self.key_counter += 1
        return new_key

    def visit(self, addr: StaticAddress):
        self.address_visitor.visit(addr)

    def yield_state(
        self,
    ) -> tuple[
        Weight,
        AddressVisitor,
        list[Trace[Any]],
    ]:
        return (
            self.weight,
            self.address_visitor,
            self.address_traces,
        )

    def get_subconstraint(
        self,
        addr: StaticAddress,
    ) -> Constraint:
        return ChoiceMapConstraint(self.choice_map(addr))

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        self.visit(addr)
        subconstraint = self.get_subconstraint(addr)
        sub_key = self.fresh_key_and_increment()
        (tr, w) = gen_fn.generate(sub_key, subconstraint, args)
        self.weight += w
        self.address_traces.append(tr)

        return tr.get_retval()


def generate_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(
        key: PRNGKey,
        choice_map: ChoiceMap,
        args: tuple[Any, ...],
    ):
        stateful_handler = GenerateHandler(key, choice_map)
        retval = forward(source_fn)(stateful_handler, *args)
        (
            weight,
            address_visitor,
            address_traces,
        ) = stateful_handler.yield_state()
        return (
            weight,
            # Trace.
            (
                args,
                retval,
                address_visitor,
                address_traces,
            ),
        )

    return wrapper


###############
# Update edit #
###############


@dataclass
class UpdateHandler(StaticHandler):
    key: PRNGKey
    previous_trace: StaticTrace[Any]
    constraint: ChoiceMap
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)
    weight: FloatArray = Pytree.field(default_factory=lambda: jnp.zeros(()))
    address_traces: list[Trace[Any]] = Pytree.field(default_factory=list)
    bwd_constraints: list[ChoiceMap] = Pytree.field(default_factory=list)
    key_counter: int = Pytree.static(default=1)

    def fresh_key_and_increment(self):
        new_key = jax.random.fold_in(self.key, self.key_counter)
        self.key_counter += 1
        return new_key

    def yield_state(self):
        return (
            self.weight,
            self.address_visitor,
            self.address_traces,
            self.bwd_constraints,
        )

    def visit(self, addr):
        self.address_visitor.visit(addr)

    def get_subconstraint(self, addr: StaticAddress) -> ChoiceMap:
        return self.constraint(addr)

    def get_subtrace(
        self,
        addr: StaticAddress,
    ):
        return self.previous_trace.get_subtrace(addr)

    def handle_retval(self, v):
        return jtu.tree_leaves(v, is_leaf=lambda v: isinstance(v, Diff))

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        argdiffs: Argdiffs = args
        self.visit(addr)
        subtrace = self.get_subtrace(addr)
        constraint = self.get_subconstraint(addr)
        sub_key = self.fresh_key_and_increment()
        request = Update(constraint)
        (tr, w, retval_diff, bwd_request) = request.edit(
            sub_key,
            subtrace,
            argdiffs,
        )
        assert isinstance(bwd_request, Update) and isinstance(
            bwd_request.constraint, ChoiceMap
        )
        self.bwd_constraints.append(bwd_request.constraint)
        self.weight += w
        self.address_traces.append(tr)

        return retval_diff


def update_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(
        key: PRNGKey,
        previous_trace: StaticTrace[R],
        constraint: ChoiceMap,
        diffs: tuple[Any, ...],
    ):
        stateful_handler = UpdateHandler(key, previous_trace, constraint)
        diff_primals = Diff.tree_primal(diffs)
        diff_tangents = Diff.tree_tangent(diffs)
        retval_diffs = incremental(source_fn)(
            stateful_handler, diff_primals, diff_tangents
        )
        retval_primals = Diff.tree_primal(retval_diffs)
        (
            weight,
            address_visitor,
            address_traces,
            bwd_requests,
        ) = stateful_handler.yield_state()
        return (
            (
                retval_diffs,
                weight,
                # Trace.
                (
                    diff_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                # Backward update problem.
                bwd_requests,
            ),
        )

    return wrapper


###################################
# Choice map edit request handler #
###################################


@dataclass
class StaticEditRequestHandler(StaticHandler):
    key: PRNGKey
    previous_trace: StaticTrace[Any]
    addressed: StaticDict
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)
    weight: FloatArray = Pytree.field(default_factory=lambda: jnp.zeros(()))
    address_traces: list[Trace[Any]] = Pytree.field(default_factory=list)
    bwd_requests: list[EditRequest] = Pytree.field(default_factory=list)
    key_counter: int = Pytree.static(default=1)

    def fresh_key_and_increment(self):
        new_key = jax.random.fold_in(self.key, self.key_counter)
        self.key_counter += 1
        return new_key

    def yield_state(self):
        return (
            self.weight,
            self.address_visitor,
            self.address_traces,
            self.bwd_requests,
        )

    def visit(self, addr):
        self.address_visitor.visit(addr)

    def get_subrequest(
        self, addr: StaticAddressComponent | StaticAddress
    ) -> EditRequest:
        addr = collapse_address(addr)
        return self.addressed.get(addr, EmptyRequest())

    def get_subtrace(
        self,
        addr: StaticAddress,
    ):
        return self.previous_trace.get_subtrace(addr)

    def handle_retval(self, v):
        return jtu.tree_leaves(v, is_leaf=lambda v: isinstance(v, Diff))

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        argdiffs: Argdiffs = args
        self.visit(addr)
        subtrace = self.get_subtrace(addr)
        subrequest = self.get_subrequest(addr)
        sub_key = self.fresh_key_and_increment()
        (tr, w, retval_diff, bwd_request) = subrequest.edit(
            sub_key,
            subtrace,
            argdiffs,
        )
        self.bwd_requests.append(bwd_request)
        self.weight += w
        self.address_traces.append(tr)

        return retval_diff


def static_edit_request_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(
        key: PRNGKey,
        previous_trace: StaticTrace[R],
        addressed: dict[StaticAddressComponent | StaticAddress, EditRequest],
        diffs: tuple[Any, ...],
    ):
        stateful_handler = StaticEditRequestHandler(
            key,
            previous_trace,
            addressed,
        )
        diff_primals = Diff.tree_primal(diffs)
        diff_tangents = Diff.tree_tangent(diffs)
        retval_diffs = incremental(source_fn)(
            stateful_handler, diff_primals, diff_tangents
        )
        retval_primals = Diff.tree_primal(retval_diffs)
        (
            weight,
            address_visitor,
            address_traces,
            bwd_requests,
        ) = stateful_handler.yield_state()
        return (
            (
                retval_diffs,
                weight,
                # Trace.
                (
                    diff_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                # Backward update problem.
                bwd_requests,
            ),
        )

    return wrapper


#####################
# Select apply edit #
#####################


@dataclass
class RegenerateRequestHandler(StaticHandler):
    key: PRNGKey
    previous_trace: StaticTrace[Any]
    selection: Selection
    edit_request: EditRequest
    address_visitor: AddressVisitor = Pytree.field(default_factory=AddressVisitor)
    weight: FloatArray = Pytree.field(default_factory=lambda: jnp.zeros(()))
    address_traces: list[Trace[Any]] = Pytree.field(default_factory=list)
    bwd_requests: list[EditRequest] = Pytree.field(default_factory=list)
    key_counter: int = Pytree.static(default=1)

    def fresh_key_and_increment(self):
        new_key = jax.random.fold_in(self.key, self.key_counter)
        self.key_counter += 1
        return new_key

    def yield_state(self):
        return (
            self.weight,
            self.address_visitor,
            self.address_traces,
            self.bwd_requests,
        )

    def visit(self, addr):
        self.address_visitor.visit(addr)

    def get_subselection(self, addr: StaticAddress) -> Selection:
        return self.selection(addr)

    def get_subtrace(
        self,
        addr: StaticAddress,
    ):
        return self.previous_trace.get_subtrace(addr)

    def handle_retval(self, v):
        return jtu.tree_leaves(v, is_leaf=lambda v: isinstance(v, Diff))

    def handle_trace(
        self,
        addr: StaticAddress,
        gen_fn: GenerativeFunction[Any],
        args: tuple[Any, ...],
    ):
        argdiffs: Argdiffs = args
        self.visit(addr)
        subtrace = self.get_subtrace(addr)
        subselection = self.get_subselection(addr)
        sub_key = self.fresh_key_and_increment()
        subrequest = Regenerate(subselection)
        tr, w, retval_diff, bwd_request = subrequest.edit(sub_key, subtrace, argdiffs)
        self.bwd_requests.append(bwd_request)
        self.weight += w
        self.address_traces.append(tr)

        return retval_diff


def regenerate_transform(source_fn):
    @functools.wraps(source_fn)
    def wrapper(
        key: PRNGKey,
        previous_trace: StaticTrace[R],
        selection: Selection,
        edit_request: EditRequest,
        diffs: tuple[Any, ...],
    ):
        stateful_handler = RegenerateRequestHandler(
            key,
            previous_trace,
            selection,
            edit_request,
        )
        diff_primals = Diff.tree_primal(diffs)
        diff_tangents = Diff.tree_tangent(diffs)
        retval_diffs = incremental(source_fn)(
            stateful_handler, diff_primals, diff_tangents
        )
        retval_primals = Diff.tree_primal(retval_diffs)
        (
            weight,
            address_visitor,
            address_traces,
            bwd_requests,
        ) = stateful_handler.yield_state()
        return (
            (
                retval_diffs,
                weight,
                # Trace.
                (
                    diff_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                # Backward update problem.
                bwd_requests,
            ),
        )

    return wrapper


#######################
# Generative function #
#######################


# Callee syntactic sugar handler.


def handler_trace_with_static(
    addr: StaticAddressComponent | StaticAddress,
    gen_fn: GenerativeFunction[Any],
    args: tuple[Any, ...],
):
    return trace(addr if isinstance(addr, tuple) else (addr,), gen_fn, args)


@Pytree.dataclass
class StaticGenerativeFunction(Generic[R], GenerativeFunction[R]):
    """A `StaticGenerativeFunction` is a generative function which relies on program
    transformations applied to JAX-compatible Python programs to implement the generative
    function interface.

    By virtue of the implementation, any source program which is provided to this generative function *must* be JAX traceable, meaning [all the footguns for programs that JAX exposes](https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html) apply to the source program.

    **Language restrictions**

    In addition to JAX footguns, there are a few more which are specific to the generative function interface semantics. Here is the full list of language restrictions (and capabilities):

    * One is allowed to use `jax.lax` control flow primitives _so long as the functions provided to the primitives do not contain `trace` invocations_. In other words, utilizing control flow primitives within the source of a `StaticGenerativeFunction`'s source program requires that the control flow primitives get *deterministic* computation.

    * The above restriction also applies to `jax.vmap`.

    * Source programs are allowed to utilize untraced randomness, although there are restrictions (which we discuss below). It is required to use [`jax.random`](https://jax.readthedocs.io/en/latest/jax.random.html) and JAX's PRNG capabilities. To utilize untraced randomness, you'll need to pass in an extra key as an argument to your model.

        ```python
        @gen
        def model(key: PRNGKey):
            v = some_untraced_call(key)
            x = trace("x", genjax.normal)(v, 1.0)
            return x
        ```
    """

    source: Closure[R]
    """
    The source program of the generative function. This is a JAX-compatible Python program.
    """

    def __get__(self, instance, _klass) -> "StaticGenerativeFunction[R]":
        """
        This method allows the @genjax.gen decorator to transform instance methods, turning them into `StaticGenerativeFunction[R]` calls.

        NOTE: if you assign an already-created `StaticGenerativeFunction` to a variable inside of a class, it will always receive the instance as its first method.
        """
        return self.partial_apply(instance) if instance else self

    # To get the type of return value, just invoke
    # the source (with abstract tracer arguments).
    def __abstract_call__(self, *args) -> Any:
        return to_shape_fn(self.source, jnp.zeros)(*args)

    def __post_init__(self):
        wrapped = self.source.fn
        # Preserve the original function's docstring and name
        for k in _WRAPPER_ASSIGNMENTS:
            v = getattr(wrapped, k, None)
            if v is not None:
                object.__setattr__(self, k, v)

        object.__setattr__(self, "__wrapped__", wrapped)

    def handle_kwargs(self) -> "StaticGenerativeFunction[R]":
        @Pytree.partial()
        def kwarged_source(args, kwargs):
            return self.source(*args, **kwargs)

        return StaticGenerativeFunction(kwarged_source)

    def simulate(
        self,
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> StaticTrace[R]:
        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )
        (args, retval, address_visitor, address_traces) = simulate_transform(
            syntax_sugar_handled
        )(key, args)
        return StaticTrace(
            self,
            args,
            retval,
            address_visitor,
            address_traces,
        )

    def generate(
        self,
        key: PRNGKey,
        constraint: Constraint,
        args: tuple[Any, ...],
    ) -> tuple[StaticTrace[R], Weight]:
        assert isinstance(constraint, ChoiceMapConstraint), type(constraint)

        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )

        (
            weight,
            # Trace.
            (
                args,
                retval,
                address_visitor,
                address_traces,
            ),
        ) = generate_transform(syntax_sugar_handled)(key, constraint.choice_map, args)
        return StaticTrace(
            self,
            args,
            retval,
            address_visitor,
            address_traces,
        ), weight

    def project(
        self,
        key: PRNGKey,
        trace: Trace[Any],
        projection: Projection[ChoiceMap],
    ) -> Weight:
        assert isinstance(trace, StaticTrace)
        assert isinstance(projection, Selection), type(projection)
        weight = jnp.array(0.0)
        for addr in trace.addresses.get_visited():
            subprojection = projection(addr)
            subtrace = trace.get_subtrace(addr)
            weight += subtrace.project(key, subprojection)
        return weight

    def edit_update(
        self,
        key: PRNGKey,
        trace: StaticTrace[R],
        constraint: ChoiceMap,
        argdiffs: Argdiffs,
    ) -> tuple[StaticTrace[R], Weight, Retdiff[R], EditRequest]:
        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )
        (
            (
                retval_diffs,
                weight,
                (
                    arg_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                bwd_requests,
            ),
        ) = update_transform(syntax_sugar_handled)(key, trace, constraint, argdiffs)
        if not Diff.static_check_tree_diff(retval_diffs):
            retval_diffs = Diff.no_change(retval_diffs)

        def make_bwd_request(visitor, subconstraints):
            addresses = visitor.get_visited()
            addresses = Pytree.tree_const_unwrap(addresses)
            chm = ChoiceMap.from_mapping(zip(addresses, subconstraints))
            return Update(chm)

        bwd_request = make_bwd_request(address_visitor, bwd_requests)
        return (
            StaticTrace(
                self,
                arg_primals,
                retval_primals,
                address_visitor,
                address_traces,
            ),
            weight,
            retval_diffs,
            bwd_request,
        )

    def edit_static_edit_request(
        self,
        key: PRNGKey,
        trace: StaticTrace[R],
        addressed: StaticDict,
        argdiffs: Argdiffs,
    ) -> tuple[StaticTrace[R], Weight, Retdiff[R], EditRequest]:
        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )
        (
            (
                retval_diffs,
                weight,
                (
                    arg_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                bwd_requests,
            ),
        ) = static_edit_request_transform(syntax_sugar_handled)(
            key, trace, addressed, argdiffs
        )

        def make_bwd_request(
            visitor: AddressVisitor,
            subrequests: list[EditRequest],
        ):
            addresses = visitor.get_visited()
            addresses = Pytree.tree_const_unwrap(addresses)
            addresses = map(collapse_address, addresses)
            bwd_addressed = dict(zip(addresses, subrequests))
            return StaticRequest(bwd_addressed)

        bwd_request = make_bwd_request(address_visitor, bwd_requests)
        return (
            StaticTrace(
                self,
                arg_primals,
                retval_primals,
                address_visitor,
                address_traces,
            ),
            weight,
            retval_diffs,
            bwd_request,
        )

    def edit_regenerate(
        self,
        key: PRNGKey,
        trace: StaticTrace[R],
        selection: Selection,
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[StaticTrace[R], Weight, Retdiff[R], EditRequest]:
        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )
        (
            (
                retval_diffs,
                weight,
                (
                    arg_primals,
                    retval_primals,
                    address_visitor,
                    address_traces,
                ),
                bwd_requests,
            ),
        ) = regenerate_transform(syntax_sugar_handled)(
            key, trace, selection, edit_request, argdiffs
        )

        def make_bwd_request(
            visitor: AddressVisitor,
            subrequests: list[EditRequest],
        ):
            addresses = visitor.get_visited()
            addresses = Pytree.tree_const_unwrap(addresses)
            addresses = map(collapse_address, addresses)
            bwd_addressed = dict(zip(addresses, subrequests))
            return StaticRequest(bwd_addressed)

        bwd_request = make_bwd_request(address_visitor, bwd_requests)
        return (
            StaticTrace(
                self,
                arg_primals,
                retval_primals,
                address_visitor,
                address_traces,
            ),
            weight,
            retval_diffs,
            bwd_request,
        )

    def edit(
        self,
        key: PRNGKey,
        trace: Trace[R],
        edit_request: EditRequest,
        argdiffs: Argdiffs,
    ) -> tuple[StaticTrace[R], Weight, Retdiff[R], EditRequest]:
        assert isinstance(trace, StaticTrace)
        match edit_request:
            case Update(constraint):
                return self.edit_update(
                    key,
                    trace,
                    constraint,
                    argdiffs,
                )

            case StaticRequest(addressed):
                return self.edit_static_edit_request(
                    key,
                    trace,
                    addressed,
                    argdiffs,
                )
            case Regenerate(selection):
                return self.edit_regenerate(
                    key,
                    trace,
                    selection,
                    edit_request,
                    argdiffs,
                )
            case _:
                raise NotSupportedEditRequest(edit_request)

    def assess(
        self,
        sample: ChoiceMap,
        args: tuple[Any, ...],
    ) -> tuple[Score, R]:
        syntax_sugar_handled = push_trace_overload_stack(
            handler_trace_with_static, self.source
        )
        (retval, score) = assess_transform(syntax_sugar_handled)(sample, args)
        return (score, retval)

    def inline(self, *args):
        return self.source(*args)

    @property
    def partial_args(self) -> tuple[Any, ...]:
        """
        Returns the partially applied arguments of the generative function.

        This method retrieves the dynamically applied arguments that were used to create
        this StaticGenerativeFunction instance through partial application.

        Returns:
            tuple[Any, ...]: A tuple containing the partially applied arguments.

        Note:
            This method is particularly useful when working with partially applied
            generative functions, allowing access to the pre-filled arguments.
        """
        return self.source.dyn_args

    def partial_apply(self, *args) -> "StaticGenerativeFunction[R]":
        """
        Returns a new [`StaticGenerativeFunction`][] with the given arguments partially applied.

        This method creates a new [`StaticGenerativeFunction`][] that has some of its arguments pre-filled. When called, the new function will use the pre-filled arguments along with any additional arguments provided.

        Args:
            *args: Variable length argument list to be partially applied to the function.

        Returns:
            A new [`StaticGenerativeFunction`][] with partially applied arguments.

        Example:
            ```python
            @gen
            def my_model(x, y):
                z = normal(x, 1.0) @ "z"
                return y * z


            partially_applied_model = my_model.partial_apply(2.0)
            # Now `partially_applied_model` is equivalent to a model that only takes 'y' as an argument
            ```
        """
        all_args = self.source.dyn_args + args
        return gen(Closure[R](all_args, self.source.fn))


#############
# Decorator #
#############


def gen(f: Closure[R] | Callable[..., R]) -> StaticGenerativeFunction[R]:
    if isinstance(f, Closure):
        return StaticGenerativeFunction[R](f)
    else:
        closure = Closure[R]((), f)
        return gen(closure)


###########
# Exports #
###########

__all__ = [
    "AddressReuse",
    "StaticGenerativeFunction",
    "gen",
    "trace",
    "trace_p",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sequential Monte Carlo ([Chopin & Papaspiliopoulos, 2020](https://link.springer.com/book/10.1007/978-3-030-47845-2), [Del Moral, Doucet, & Jasram 2006](https://academic.oup.com/jrsssb/article/68/3/411/7110641)) is an approximate inference framework based on approximating a sequence of target distributions using a weighted collection of particles.

In this module, we provide a set of ingredients for implementing SMC algorithms, including pseudomarginal / recursive auxiliary variants, and variants expressible using SMCP3 ([Lew & Matheos, et al, 2024](https://proceedings.mlr.press/v206/lew23a/lew23a.pdf)) moves.
"""

from abc import abstractmethod

from jax import numpy as jnp
from jax import random as jrandom
from jax import tree_util as jtu
from jax import vmap
from jax.scipy.special import logsumexp

from genjax._src.core.generative import (
    ChoiceMap,
    Trace,
)
from genjax._src.core.generative.core import Score
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    ArrayLike,
    BoolArray,
    FloatArray,
    Generic,
    PRNGKey,
    TypeVar,
)
from genjax._src.generative_functions.distributions.tensorflow_probability import (
    categorical,
)
from genjax._src.inference.sp import (
    Algorithm,
    SampleDistribution,
    Target,
)

R = TypeVar("R")

# Utility, for CSMC stacking.


def stack_to_first_dim(arr1: ArrayLike, arr2: ArrayLike):
    # Coerce to array, if literal.
    arr1 = jnp.array(arr1, copy=False)
    arr2 = jnp.array(arr2, copy=False)
    # Ensure both arrays are at least 2D
    if arr1.ndim <= 1:
        arr1 = arr1.reshape(-1, 1)
    if arr2.ndim <= 1:
        arr2 = arr2.reshape(-1, 1)

    # Stack the arrays along the first dimension
    result = jnp.concatenate([arr1, arr2], axis=0)
    return jnp.squeeze(result)


#######################
# Particle collection #
#######################


@Pytree.dataclass
class ParticleCollection(Generic[R], Pytree):
    """A collection of weighted particles.

    Stores the particles (which are `Trace` instances), the log importance weights, the log marginal likelihood estimate, as well as an indicator flag denoting whether the collection is runtime valid or not (`ParticleCollection.is_valid`).
    """

    particles: Trace[R]
    log_weights: FloatArray
    is_valid: BoolArray

    def get_particles(self) -> Trace[R]:
        return self.particles

    def get_particle(self, idx) -> Trace[R]:
        return jtu.tree_map(lambda v: v[idx], self.particles)

    def get_log_weights(self) -> FloatArray:
        return self.log_weights

    def get_log_marginal_likelihood_estimate(self) -> FloatArray:
        return logsumexp(self.log_weights) - jnp.log(len(self.log_weights))

    def __getitem__(self, idx) -> tuple[Any, ...]:
        return jtu.tree_map(lambda v: v[idx], (self.particles, self.log_weights))

    def sample_particle(self, key) -> Trace[R]:
        """
        Samples a particle from the collection, with probability proportional to its weight.
        """
        log_weights = self.get_log_weights()
        logits = log_weights - logsumexp(log_weights)
        _, idx = categorical.random_weighted(key, logits)
        return self.get_particle(idx)


####################################
# Abstract type for SMC algorithms #
####################################


class SMCAlgorithm(Generic[R], Algorithm[R]):
    """Abstract class for SMC algorithms."""

    @abstractmethod
    def get_num_particles(self) -> int:
        pass

    @abstractmethod
    def get_final_target(self) -> Target[R]:
        pass

    @abstractmethod
    def run_smc(
        self,
        key: PRNGKey,
    ) -> ParticleCollection[R]:
        pass

    @abstractmethod
    def run_csmc(
        self,
        key: PRNGKey,
        retained: ChoiceMap,
    ) -> ParticleCollection[R]:
        pass

    # Convenience method for returning an estimate of the normalizing constant
    # of the target.
    def log_marginal_likelihood_estimate(
        self,
        key: PRNGKey,
        target: Target[R] | None = None,
    ):
        if target:
            algorithm = ChangeTarget(self, target)
        else:
            algorithm = self
        key, sub_key = jrandom.split(key)
        particle_collection = algorithm.run_smc(sub_key)
        return particle_collection.get_log_marginal_likelihood_estimate()

    #########
    # GenSP #
    #########

    def random_weighted(
        self,
        key: PRNGKey,
        *args: Any,
    ) -> tuple[Score, ChoiceMap]:
        assert isinstance(args[0], Target)

        target: Target[R] = args[0]
        algorithm = ChangeTarget(self, target)
        key, sub_key = jrandom.split(key)
        particle_collection = algorithm.run_smc(key)
        particle = particle_collection.sample_particle(sub_key)
        log_density_estimate = (
            particle.get_score()
            - particle_collection.get_log_marginal_likelihood_estimate()
        )
        chm = target.filter_to_unconstrained(particle.get_sample())
        return log_density_estimate, chm

    def estimate_logpdf(
        self,
        key: PRNGKey,
        v: ChoiceMap,
        *args: tuple[Any, ...],
    ) -> Score:
        assert isinstance(args[0], Target)

        target: Target[R] = args[0]
        algorithm = ChangeTarget(self, target)
        key, sub_key = jrandom.split(key)
        particle_collection = algorithm.run_csmc(key, v)
        particle = particle_collection.sample_particle(sub_key)
        log_density_estimate = (
            particle.get_score()
            - particle_collection.get_log_marginal_likelihood_estimate()
        )
        return log_density_estimate

    ################
    # VI via GRASP #
    ################

    def estimate_normalizing_constant(
        self,
        key: PRNGKey,
        target: Target[R],
    ) -> FloatArray:
        algorithm = ChangeTarget(self, target)
        key, sub_key = jrandom.split(key)
        particle_collection = algorithm.run_smc(sub_key)
        return particle_collection.get_log_marginal_likelihood_estimate()

    def estimate_reciprocal_normalizing_constant(
        self,
        key: PRNGKey,
        target: Target[R],
        latent_choices: ChoiceMap,
        w: FloatArray,
    ) -> FloatArray:
        algorithm = ChangeTarget(self, target)
        # Special, for ChangeTarget -- to avoid a redundant reweighting step,
        # when we have `w` which (with `latent_choices`) is already properly weighted
        # for the `target`.
        return algorithm.run_csmc_for_normalizing_constant(key, latent_choices, w)


#######################
# Importance sampling #
#######################


@Pytree.dataclass
class Importance(Generic[R], SMCAlgorithm[R]):
    """Accepts as input a `target: Target` and, optionally, a proposal `q: SampleDistribution`.
    `q` should accept a `Target` as input and return a choicemap on a subset
    of the addresses in `target.gen_fn` not in `target.constraints`.

    This initializes a 1-particle `ParticleCollection` by importance sampling from `target` using `q`.

    Any choices in `target.p` not in `q` will be sampled from the internal proposal distribution of `p`,
    given `target.constraints` and the choices sampled by `q`.
    """

    target: Target[R]
    q: SampleDistribution | None = Pytree.field(default=None)

    def get_num_particles(self):
        return 1

    def get_final_target(self):
        return self.target

    def run_smc(self, key: PRNGKey):
        key, sub_key = jrandom.split(key)
        if self.q is not None:
            log_weight, choice = self.q.random_weighted(sub_key, self.target)
            tr, target_score = self.target.importance(key, choice)
        else:
            log_weight = 0.0
            tr, target_score = self.target.importance(key, ChoiceMap.empty())
        return ParticleCollection(
            jtu.tree_map(lambda v: jnp.expand_dims(v, axis=0), tr),
            jnp.array([target_score - log_weight]),
            jnp.array(True),
        )

    def run_csmc(self, key: PRNGKey, retained: ChoiceMap):
        key, sub_key = jrandom.split(key)
        if self.q:
            q_score = self.q.estimate_logpdf(sub_key, retained, self.target)
        else:
            q_score = 0.0
        target_trace, target_score = self.target.importance(key, retained)
        return ParticleCollection(
            jtu.tree_map(lambda v: jnp.expand_dims(v, axis=0), target_trace),
            jnp.array([target_score - q_score]),
            jnp.array(True),
        )


@Pytree.dataclass
class ImportanceK(Generic[R], SMCAlgorithm[R]):
    """Given a `target: Target` and a proposal `q: SampleDistribution`, as well as the
    number of particles `k_particles: int`, initialize a particle collection using
    importance sampling."""

    target: Target[R]
    q: SampleDistribution | None = Pytree.field(default=None)
    k_particles: int = Pytree.static(default=2)

    def get_num_particles(self):
        return self.k_particles

    def get_final_target(self):
        return self.target

    def run_smc(self, key: PRNGKey):
        key, sub_key = jrandom.split(key)
        sub_keys = jrandom.split(sub_key, self.get_num_particles())
        if self.q is not None:
            log_weights, choices = vmap(self.q.random_weighted, in_axes=(0, None))(
                sub_keys, self.target
            )
            trs, target_scores = vmap(self.target.importance)(sub_keys, choices)
        else:
            log_weights = 0.0
            trs, target_scores = vmap(self.target.importance, in_axes=(0, None))(
                sub_keys, ChoiceMap.empty()
            )
        return ParticleCollection(
            trs,
            target_scores - log_weights,
            jnp.array(True),
        )

    def run_csmc(self, key: PRNGKey, retained: ChoiceMap):
        key, sub_key = jrandom.split(key)
        sub_keys = jrandom.split(sub_key, self.get_num_particles() - 1)
        if self.q:
            log_scores, choices = vmap(self.q.random_weighted, in_axes=(0, None))(
                sub_keys, self.target
            )
            retained_choice_score = self.q.estimate_logpdf(key, retained, self.target)
            stacked_choices = jtu.tree_map(stack_to_first_dim, choices, retained)
            stacked_scores = jtu.tree_map(
                stack_to_first_dim, log_scores, retained_choice_score
            )
            sub_keys = jrandom.split(key, self.get_num_particles())
            target_traces, target_scores = vmap(self.target.importance)(
                sub_keys, stacked_choices
            )
        else:
            ignored_traces, ignored_scores = vmap(
                self.target.importance, in_axes=(0, None)
            )(sub_keys, ChoiceMap.empty())
            retained_trace, retained_choice_score = self.target.importance(
                key, retained
            )
            target_scores = jtu.tree_map(
                stack_to_first_dim, ignored_scores, retained_choice_score
            )
            stacked_scores = 0.0
            target_traces = jtu.tree_map(
                stack_to_first_dim, ignored_traces, retained_trace
            )
        return ParticleCollection(
            target_traces,
            target_scores - stacked_scores,
            jnp.array(True),
        )


#################
# Change target #
#################


@Pytree.dataclass
class ChangeTarget(Generic[R], SMCAlgorithm[R]):
    prev: SMCAlgorithm[R]
    target: Target[R]

    def get_num_particles(self):
        return self.prev.get_num_particles()

    def get_final_target(self):
        return self.target

    def run_smc(
        self,
        key: PRNGKey,
    ) -> ParticleCollection[R]:
        collection = self.prev.run_smc(key)

        # Convert the existing set of particles and weights
        # to a new set which is properly weighted for the new target.
        def _reweight(key, particle, weight) -> tuple[Trace[R], Any]:
            latents = self.prev.get_final_target().filter_to_unconstrained(
                particle.get_sample()
            )
            new_trace, new_weight = self.target.importance(key, latents)
            this_weight = new_weight - particle.get_score() + weight
            return (new_trace, this_weight)

        sub_keys = jrandom.split(key, self.get_num_particles())
        new_particles, new_weights = vmap(_reweight)(
            sub_keys,
            collection.get_particles(),
            collection.get_log_weights(),
        )
        return ParticleCollection(
            new_particles,
            new_weights,
            jnp.array(True),
        )

    def run_csmc(
        self,
        key: PRNGKey,
        retained: ChoiceMap,
    ) -> ParticleCollection[R]:
        collection = self.prev.run_csmc(key, retained)

        # Convert the existing set of particles and weights
        # to a new set which is properly weighted for the new target.
        def _reweight(key, particle, weight) -> tuple[Trace[R], Any]:
            latents = self.prev.get_final_target().filter_to_unconstrained(
                particle.get_sample()
            )
            new_trace, new_score = self.target.importance(key, latents)
            this_weight = new_score - particle.get_score() + weight
            return (new_trace, this_weight)

        sub_keys = jrandom.split(key, self.get_num_particles())
        new_particles, new_weights = vmap(_reweight)(
            sub_keys,
            collection.get_particles(),
            collection.get_log_weights(),
        )
        return ParticleCollection(
            new_particles,
            new_weights,
            jnp.array(True),
        )

    # NOTE: This method is specialized to support the variational inference interface
    # `estimate_reciprocal_normalizing_constant` - by avoiding an extra target
    # reweighting step (which will add extra variance to any derived gradient estimators)
    # It is only available for `ChangeTarget`.

    def run_csmc_for_normalizing_constant(
        self,
        key: PRNGKey,
        latent_choices: ChoiceMap,
        w: FloatArray,
    ) -> FloatArray:
        key, sub_key = jrandom.split(key)
        particle_collection = self.prev.run_csmc(sub_key, latent_choices)

        # Convert the existing set of particles and weights
        # to a new set which is properly weighted for the new target.
        def _reweight(key, particle, weight):
            latents = self.prev.get_final_target().filter_to_unconstrained(
                particle.get_sample()
            )
            _, new_score = self.target.importance(key, latents)
            this_weight = new_score - particle.get_score() + weight
            return this_weight

        num_particles = self.get_num_particles()
        sub_keys = jrandom.split(key, num_particles - 1)
        new_rejected_weights = vmap(_reweight)(
            sub_keys,
            jtu.tree_map(lambda v: v[:-1], particle_collection.get_particles()),
            jtu.tree_map(lambda v: v[:-1], particle_collection.get_log_weights()),
        )
        retained_score = particle_collection.get_particle(-1).get_score()
        retained_weight = particle_collection.get_log_weights()[-1]
        all_weights = stack_to_first_dim(
            new_rejected_weights,
            w - retained_score + retained_weight,
        )
        total_weight = logsumexp(all_weights)
        return retained_score - (total_weight - jnp.log(num_particles))



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides:

* Abstract base classes for declaring distributions as `GenerativeFunction` types. These classes are `Distribution` and `ExactDensity`. The latter assumes that the inheritor exposes _exact density evaluation_, while the former makes no such assumption.

* A number of distribution generative functions which implement `ExactDensity` by utilizing TensorFlow Probability distributions with the JAX backend.

* Custom distributions, including ones with exact posteriors (like discrete HMMs).
"""



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from genjax._src.adev.core import ADEVPrimitive, Dual, expectation, sample_primitive
from genjax._src.adev.primitives import (
    add_cost,
    baseline,
    beta_implicit,
    categorical_enum_parallel,
    flip_enum,
    flip_enum_parallel,
    flip_mvd,
    flip_reinforce,
    geometric_reinforce,
    mv_normal_diag_reparam,
    mv_normal_reparam,
    normal_reinforce,
    normal_reparam,
    reinforce,
    uniform,
)

__all__ = [
    "ADEVPrimitive",
    "Dual",
    "add_cost",
    "baseline",
    "beta_implicit",
    "categorical_enum_parallel",
    "expectation",
    # Primitives.
    "flip_enum",
    "flip_enum_parallel",
    "flip_mvd",
    "flip_reinforce",
    "geometric_reinforce",
    "mv_normal_diag_reparam",
    "mv_normal_reparam",
    "normal_reinforce",
    "normal_reparam",
    "reinforce",
    # Language.
    "sample_primitive",
    "uniform",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from genjax._src.inference.sp import (
    Algorithm,
    Marginal,
    SampleDistribution,
    Target,
    marginal,
)
from genjax.inference import (
    requests,
    smc,
    vi,
)

__all__ = [
    "Algorithm",
    "Marginal",
    "SampleDistribution",
    "Target",
    "marginal",
    "requests",
    "smc",
    "vi",
]



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
import jax.random as jrand
import jax.tree_util as jtu
from jax import grad
from jax.lax import scan
from tensorflow_probability.substrates import jax as tfp

from genjax._src.core.generative import (
    Argdiffs,
    ChoiceMap,
    EditRequest,
    Retdiff,
    Score,
    Selection,
    Trace,
    Update,
    Weight,
)
from genjax._src.core.generative.requests import DiffAnnotate
from genjax._src.core.interpreters.incremental import Diff
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    FloatArray,
    IntArray,
    PRNGKey,
    static_check_supports_grad,
)

tfd = tfp.distributions


# Pytree manipulation utilities -- these handle unzipping Pytrees into
# differentiable and non-diff pieces, and then also zipping them back up.
def grad_tree_unzip(tree: ChoiceMap) -> tuple[ChoiceMap, ChoiceMap]:
    grad_tree = jtu.tree_map(
        lambda v: v if static_check_supports_grad(v) else None, tree
    )
    nongrad_tree = jtu.tree_map(
        lambda v: v if not static_check_supports_grad(v) else None, tree
    )
    return grad_tree, nongrad_tree


def grad_tree_zip(
    grad_tree: ChoiceMap,
    nongrad_tree: ChoiceMap,
) -> ChoiceMap:
    return jtu.tree_map(
        lambda v1, v2: v1 if v1 is not None else v2, grad_tree, nongrad_tree
    )


# Compute the gradient of a selection of random choices
# in a trace -- uses `GenerativeFunction.assess`.
def selection_gradient(
    selection: Selection,
    trace: Trace[Any],
    argdiffs: Argdiffs,
) -> tuple[ChoiceMap, ChoiceMap]:
    chm = trace.get_choices()
    filtered = chm.filter(selection)
    complement = chm.filter(~selection)
    grad_tree, nongrad_tree = grad_tree_unzip(filtered)
    gen_fn = trace.get_gen_fn()

    def differentiable_assess(grad_tree):
        zipped = grad_tree_zip(grad_tree, nongrad_tree)
        full_choices = zipped.merge(complement)
        weight, _ = gen_fn.assess(
            full_choices,
            Diff.tree_primal(argdiffs),
        )
        return weight

    return grad_tree_zip(grad_tree, nongrad_tree), jtu.tree_map(
        lambda v1, v2: v1
        if v1 is not None
        else jnp.zeros_like(jnp.array(v2, copy=False)),
        grad(differentiable_assess)(grad_tree),
        nongrad_tree,
    )


# Utilities for momenta sampling and score evaluation.
def normal_sample(key: PRNGKey, shape) -> FloatArray:
    return tfd.Normal(jnp.zeros(shape), 1.0).sample(seed=key)


def normal_score(v) -> Score:
    score = tfd.Normal(0.0, 1.0).log_prob(v)
    if score.shape:
        return jnp.sum(score)
    else:
        return score


def assess_momenta(momenta, mul=1.0):
    return jnp.sum(
        jnp.array(
            jtu.tree_leaves(jtu.tree_map(lambda v: normal_score(mul * v), momenta))
        )
    )


def sample_momenta(key, choice_gradients):
    total_length = len(jtu.tree_leaves(choice_gradients))
    int_seeds = jnp.arange(total_length)
    int_seed_tree = jtu.tree_unflatten(jtu.tree_structure(choice_gradients), int_seeds)
    momenta_tree = jtu.tree_map(
        lambda v, int_seed: normal_sample(jrand.fold_in(key, int_seed), v.shape),
        choice_gradients,
        int_seed_tree,
    )
    momenta_score = assess_momenta(momenta_tree)
    return momenta_tree, momenta_score


#######
# HMC #
#######


@Pytree.dataclass(match_args=True)
class HMC(EditRequest):
    """
    Apply a Hamiltonian Monte Carlo (HMC) update that proposes new values for the selected addresses,
    returning the new trace, and a weight which is equal to the alpha accept-reject ratio computation for HMC.

    Hamilton's equations are numerically integrated using leapfrog integration with step size `eps` for `L` steps.
    See equations (5.18)-(5.20) of Neal (2011).

    # References
    Neal, Radford M. (2011), "MCMC Using Hamiltonian Dynamics", Handbook of Markov Chain Monte Carlo,
    pp. 113-162. URL: http://www.mcmchandbook.net/HandbookChapter5.pdf
    """

    selection: Selection
    eps: FloatArray
    L: int = Pytree.static(default=10)

    def edit(
        self,
        key: PRNGKey,
        tr: Trace[Any],
        argdiffs: Argdiffs,
    ) -> tuple[Trace[Any], Weight, Retdiff[Any], "EditRequest"]:
        # Just a conservative restriction, for now.
        assert Diff.static_check_no_change(argdiffs)

        original_model_score = tr.get_score()
        values, gradients = selection_gradient(self.selection, tr, argdiffs)
        key, sub_key = jrand.split(key)
        momenta, original_momenta_score = sample_momenta(sub_key, gradients)

        def kernel(
            carry: tuple[Trace[Any], ChoiceMap, ChoiceMap, ChoiceMap],
            scanned_in: IntArray,
        ) -> tuple[tuple[Trace[Any], ChoiceMap, ChoiceMap, ChoiceMap], Retdiff[Any]]:
            trace, values, gradient, momenta = carry
            int_seed = scanned_in
            momenta = jtu.tree_map(
                lambda v, g: v + (self.eps / 2) * g, momenta, gradient
            )
            values = jtu.tree_map(lambda v, m: v + self.eps * m, values, momenta)
            new_key = jrand.fold_in(key, int_seed)
            new_trace, _, retdiff, _ = Update(values).edit(new_key, trace, argdiffs)
            values, gradients = selection_gradient(self.selection, new_trace, argdiffs)
            momenta = jtu.tree_map(
                lambda v, g: v + (self.eps / 2) * g, momenta, gradients
            )
            return (new_trace, values, gradient, momenta), retdiff

        int_seeds = jnp.arange(self.L) + 1
        (final_trace, _, _, final_momenta), retdiffs = scan(
            kernel,
            (tr, values, gradients, momenta),
            int_seeds,
            length=self.L,
        )

        final_model_score = final_trace.get_score()
        final_momenta_score = assess_momenta(final_momenta, mul=-1.0)
        alpha = (
            final_model_score
            - original_model_score
            + final_momenta_score
            - original_momenta_score
        )
        # Grab the last retdiff.
        retdiff = jtu.tree_map(lambda v: v[-1], retdiffs)
        return (
            final_trace,
            alpha,
            retdiff,
            HMC(self.selection, self.eps, self.L),
        )


def SafeHMC(
    selection: Selection,
    eps: FloatArray,
    L: int = 10,
) -> DiffAnnotate[HMC]:
    def retdiff_assertion(retdiff: Retdiff[Any]):
        assert Diff.static_check_no_change(retdiff)
        return retdiff

    return HMC(selection, eps, L).map(retdiff_assertion)



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

from genjax._src.adev.core import (
    ADEVPrimitive,
    expectation,
    sample_primitive,
)
from genjax._src.adev.primitives import (
    categorical_enum_parallel,
    flip_enum,
    flip_mvd,
    geometric_reinforce,
    mv_normal_diag_reparam,
    normal_reinforce,
    normal_reparam,
)
from genjax._src.core.generative import Arguments, ChoiceMap
from genjax._src.core.typing import (
    Any,
    Callable,
    FloatArray,
    PRNGKey,
)
from genjax._src.generative_functions.distributions.distribution import (
    ExactDensity,
    exact_density,
)
from genjax._src.generative_functions.distributions.tensorflow_probability import (
    flip,
    geometric,
    normal,
)
from genjax._src.inference.smc import Importance, ImportanceK
from genjax._src.inference.sp import SampleDistribution, Target

tfd = tfp.distributions


##########################################
# Differentiable distribution primitives #
##########################################


def adev_distribution(
    adev_primitive: ADEVPrimitive,
    differentiable_logpdf: Callable[..., Any],
) -> ExactDensity[Any]:
    """
    Return an [`ExactDensity`][genjax.ExactDensity] distribution whose sampler invokes an ADEV sampling primitive, with a provided differentiable log density function.

    Exact densities created using this function can be used as distributions in variational guide programs.
    """

    def sampler(key: PRNGKey, *args: Any) -> Any:
        return sample_primitive(adev_primitive, *args, key=key)

    def logpdf(v: Any, *args: Any) -> FloatArray:
        lp = differentiable_logpdf(v, *args)
        # Branching here is statically resolved.
        if lp.shape:
            return jnp.sum(lp)
        else:
            return lp

    return exact_density(sampler, logpdf)


def logpdf(gen_fn):
    return lambda v, *args: gen_fn.assess(ChoiceMap.choice(v), args)[0]


# We import ADEV specific sampling primitives, but then wrap them in
# adev_distribution, for usage inside of generative functions.
flip_enum = adev_distribution(
    flip_enum,
    logpdf(flip),
)

flip_mvd = adev_distribution(
    flip_mvd,
    logpdf(flip),
)

categorical_enum = adev_distribution(
    categorical_enum_parallel,
    lambda v, probs: tfd.Categorical(probs=probs).log_prob(v),
)

normal_reinforce = adev_distribution(
    normal_reinforce,
    logpdf(normal),
)

normal_reparam = adev_distribution(
    normal_reparam,
    logpdf(normal),
)

mv_normal_diag_reparam = adev_distribution(
    mv_normal_diag_reparam,
    lambda v, loc, scale_diag: tfd.MultivariateNormalDiag(
        loc=loc, scale_diag=scale_diag
    ).log_prob(v),
)

geometric_reinforce = adev_distribution(
    geometric_reinforce,
    logpdf(geometric),
)


##############
# Loss terms #
##############

GradientEstimate = Any
"""
The type of gradient estimates returned by sampling from gradient estimators for loss terms.
"""


def ELBO(
    guide: SampleDistribution,
    make_target: Callable[..., Target[Any]],
) -> Callable[[PRNGKey, Arguments], GradientEstimate]:
    """
    Return a function that computes the gradient estimate of the ELBO loss term.
    """

    def grad_estimate(
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> tuple[Any, ...]:
        # In the source language of ADEV.
        @expectation
        def _loss(*args):
            target = make_target(*args)
            guide_alg = Importance(target, guide)
            w = guide_alg.estimate_normalizing_constant(key, target)
            return -w

        return _loss.grad_estimate(key, args)

    return grad_estimate


def IWELBO(
    proposal: SampleDistribution,
    make_target: Callable[[Any], Target[Any]],
    N: int,
) -> Callable[[PRNGKey, Arguments], GradientEstimate]:
    """
    Return a function that computes the gradient estimate of the IWELBO loss term.
    """

    def grad_estimate(
        key: PRNGKey,
        args: Arguments,
    ) -> GradientEstimate:
        # In the source language of ADEV.
        @expectation
        def _loss(*args):
            target = make_target(*args)
            guide = ImportanceK(target, proposal, N)
            w = guide.estimate_normalizing_constant(key, target)
            return -w

        return _loss.grad_estimate(key, args)

    return grad_estimate


def PWake(
    posterior_approx: SampleDistribution,
    make_target: Callable[[Any], Target[Any]],
) -> Callable[[PRNGKey, Arguments], GradientEstimate]:
    """
    Return a function that computes the gradient estimate of the PWake loss term.
    """

    def grad_estimate(
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> tuple[Any, ...]:
        key, sub_key1, sub_key2 = jax.random.split(key, 3)

        # In the source language of ADEV.
        @expectation
        def _loss(*target_args):
            target = make_target(*target_args)
            _, sample = posterior_approx.random_weighted(sub_key1, target)
            tr, _ = target.importance(sub_key2, sample)
            return -tr.get_score()

        return _loss.grad_estimate(key, args)

    return grad_estimate


def QWake(
    proposal: SampleDistribution,
    posterior_approx: SampleDistribution,
    make_target: Callable[[Any], Target[Any]],
) -> Callable[[PRNGKey, Arguments], GradientEstimate]:
    """
    Return a function that computes the gradient estimate of the QWake loss term.
    """

    def grad_estimate(
        key: PRNGKey,
        args: tuple[Any, ...],
    ) -> tuple[Any, ...]:
        key, sub_key1, sub_key2 = jax.random.split(key, 3)

        # In the source language of ADEV.
        @expectation
        def _loss(*target_args):
            target = make_target(*target_args)
            _, sample = posterior_approx.random_weighted(sub_key1, target)
            w = proposal.estimate_logpdf(sub_key2, sample, target)
            return -w

        return _loss.grad_estimate(key, args)

    return grad_estimate



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
import numpy as np
from scipy.linalg import circulant
from tensorflow_probability.substrates import jax as tfp

from genjax._src.core.generative.core import Score
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Array,
    FloatArray,
    IntArray,
    PRNGKey,
    TypeVar,
)
from genjax._src.generative_functions.distributions.distribution import Distribution

tfd = tfp.distributions

R = TypeVar("R")

#####
# Discrete HMM configuration
#####


def scaled_circulant(N, k, epsilon, delta):
    source = [
        epsilon ** abs(index)
        if index <= k
        else epsilon ** abs(index - N)
        if index - N >= -k
        else -delta
        for index in range(0, N)
    ]
    return circulant(source)


@Pytree.dataclass
class DiscreteHMMConfiguration(Pytree):
    linear_grid_dim: IntArray = Pytree.static()
    adjacency_distance_trans: IntArray = Pytree.static()
    adjacency_distance_obs: IntArray = Pytree.static()
    sigma_trans: FloatArray = Pytree.static()
    sigma_obs: FloatArray = Pytree.static()

    @staticmethod
    def copy(config, transition_tensor, observation_tensor):
        return DiscreteHMMConfiguration(
            config.linear_grid_dim,
            config.adjacency_distance_trans,
            config.adjacency_distance_obs,
            config.sigma_trans,
            config.sigma_obs,
        )

    def transition_tensor(self):
        return scaled_circulant(
            self.linear_grid_dim,
            self.adjacency_distance_trans,
            self.sigma_trans if self.sigma_trans > 0.0 else -np.inf,
            1 / self.sigma_trans if self.sigma_trans > 0.0 else -np.inf,
        )

    def observation_tensor(self):
        return scaled_circulant(
            self.linear_grid_dim,
            self.adjacency_distance_obs,
            self.sigma_obs if self.sigma_obs > 0.0 else -np.inf,
            1 / self.sigma_obs if self.sigma_obs > 0.0 else np.inf,
        )


#####
# Forward-filtering backward sampling
#####

# This implements JAX compatible forward-filtering
# backward sampling (to produce exact samples from discrete HMM
# posteriors)


def forward_filtering_backward_sampling(
    key: PRNGKey, config: DiscreteHMMConfiguration, observation_sequence
):
    init = int(config.linear_grid_dim / 2)
    tt = config.transition_tensor()
    prior = jnp.log(jax.nn.softmax(tt[init, :]))
    transition_n = jnp.log(jax.nn.softmax(tt))
    obs_n = jnp.log(jax.nn.softmax(config.observation_tensor()))

    # Computing the alphas and forward filter distributions:
    #
    # \alpha_1(x_1) = p(x_1) * p(y_1 | x_1) [[ initialization ]]
    #
    # \alpha_t(x_t) = p(y_t | x_t) * \sum_{x_{t-1}=1}^N
    #           [ p(x_{t-1}, y_1, ..., y_{t-1})* p(x_t | x_{t-1}) ]
    #
    #               = p(y_t | x_t)
    #           * \sum_{x_{t-1}=1}^N \alpha_t(x_t) * p(x_t | x_{t-1})
    #                                for t=2, .., T

    def forward_pass(carry, x):
        index, prev = carry
        obs = x

        def t_branch(prev, obs):
            alpha = jax.scipy.special.logsumexp(
                prev + transition_n,
                axis=-1,
            )
            alpha = obs_n + alpha.reshape(-1, 1)
            alpha = alpha[:, obs]
            return alpha

        def init_branch(prev, obs):
            alpha = obs_n + prev.reshape(-1, 1)
            alpha = alpha[:, obs]
            return alpha

        check = index == 0
        alpha = jax.lax.cond(check, init_branch, t_branch, prev, obs)
        forward_filter = alpha - jax.scipy.special.logsumexp(alpha)
        return (index + 1, alpha), (alpha, forward_filter)

    _, (_alpha, forward_filters) = jax.lax.scan(
        forward_pass, (0, prior), observation_sequence
    )

    # Computing the backward distributions.
    # Start by sampling from x_T from p(x_T | y_1, .., y_T)
    # which is the last forward filter distribution.
    #
    # Then:
    # p(x_{t-1} | x_t, y_{1:T}) = p(x_{t-1} | y_{1:t-1})
    #                                   * p(x_t | x_{t-1}) / Z_t
    #
    # where Z_t = \sum_{x_{t-1}=1}^N p(x_{t-1} | y_{1:t-1})
    #                                   * p(x_t | x_{t-1})

    def backward_sample(carry, x):
        key, index, prev_sample = carry
        forward_filter = x

        def end_branch(key, prev, forward_filter):
            sample = jax.random.categorical(key, forward_filter)
            return sample

        def t_1_branch(key, prev, forward_filter):
            backward_distribution = forward_filter + transition_n[:, prev_sample]
            backward_distribution = backward_distribution - jax.scipy.special.logsumexp(
                backward_distribution
            )
            sample = jax.random.categorical(key, backward_distribution)
            return sample

        key, sub_key = jax.random.split(key)
        check = index == 0
        sample = jax.lax.cond(
            check,
            end_branch,
            t_1_branch,
            sub_key,
            prev_sample,
            forward_filter,
        )
        return (key, index + 1, sample), sample

    # This is supposed to be scanned in reverse order
    # from the forward order.
    (key, _, _), samples = jax.lax.scan(
        backward_sample,
        (key, 0, 0),
        jnp.flip(forward_filters, axis=0),
    )
    samples = jnp.flip(samples)
    return key, (samples, forward_filters)


#####
# Exact latent sequence posterior
#####


def latent_marginals(config: DiscreteHMMConfiguration, observation_sequence):
    init = int(config.linear_grid_dim / 2)
    initial_distribution = tfd.Categorical(logits=config.transition_tensor()[init, :])
    transition_distribution = tfd.Categorical(logits=config.transition_tensor)
    observation_distribution = tfd.Categorical(logits=config.observation_tensor)
    hmm = tfd.HiddenMarkovModel(
        initial_distribution,
        transition_distribution,
        observation_distribution,
        len(observation_sequence),
    )
    marginals = hmm.posterior_marginals(observation_sequence)
    return hmm, marginals


def log_data_marginal(config, observation_sequence):
    hmm, _ = latent_marginals(config, observation_sequence)
    return hmm.log_prob(observation_sequence)


def latent_sequence_posterior(
    config: DiscreteHMMConfiguration, latent_point, observation_sequence
):
    hmm, _ = latent_marginals(config, observation_sequence)

    def _inner(carry, x):
        latent, obs = x
        v = jnp.log(carry[latent])
        v += jnp.log(jax.nn.softmax(hmm.observation_distribution.logits)[latent, obs])
        carry = jax.nn.softmax(hmm.transition_distribution.logits[latent, :])
        return carry, v

    _, probs = jax.lax.scan(
        _inner,
        jax.nn.softmax(hmm.initial_distribution.logits),
        (latent_point, observation_sequence),
    )
    prod = jnp.sum(probs)
    prod -= hmm.log_prob(observation_sequence)
    return prod, (probs, hmm.log_prob(observation_sequence))


@Pytree.dataclass
class _DiscreteHMMLatentSequencePosterior(Distribution[Array]):
    def random_weighted(self, key, *args, **kwargs) -> tuple[Score, Array]:
        config, observation_sequence = args
        key, k1, k2 = jax.random.split(key, 3)
        _, (v, _) = forward_filtering_backward_sampling(
            k1, config, observation_sequence
        )

        w = self.estimate_logpdf(k2, v, config, observation_sequence, **kwargs)
        return (w, v)

    def estimate_logpdf(self, key, v, *args, **kwargs) -> Array:
        config, observation_sequence = args
        prob, _ = latent_sequence_posterior(config, v, observation_sequence)
        return prob

    def data_logpdf(self, config, observation_sequence):
        return log_data_marginal(config, observation_sequence)


##############
# Shorthands #
##############

DiscreteHMM = _DiscreteHMMLatentSequencePosterior()



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GenJAX is a probabilistic programming system constructed by combining the concepts of
Gen with the program transformation and hardware accelerator compilation capabilities of
JAX."""

# This __init__ file exports GenJAX's public API.
# For the internals, see _src.

from importlib import metadata

from beartype import BeartypeConf
from beartype.claw import beartype_this_package

conf = BeartypeConf(
    is_color=True,
    is_debug=False,
    is_pep484_tower=True,
    violation_type=TypeError,
)

beartype_this_package(conf=conf)

from .checkify import *
from .core import *
from .experimental import *
from .generative_functions import *
from .incremental import *
from .inference import *
from .pretty import *

__version__ = metadata.version("genjax")



# Copyright 2024 The MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module supports incremental computation using a form of JVP-inspired computation
with a type of generalized tangent values (e.g. `ChangeTangent` below).

Incremental computation is currently a concern of Gen's `edit` GFI method - and can be utilized _as a runtime performance optimization_ for computing the weight (and changes to `Trace` instances) which `edit` computes.

*Change types*

By default, `genjax` provides two types of `ChangeTangent`:

* `NoChange` - indicating that a value has not changed.
* `UnknownChange` - indicating that a value has changed, without further information about the change.

`ChangeTangents` are provided along with primal values into `Diff` instances. The generative function `edit` interface expects tuples of `Pytree` instances whose leaves are `Diff` instances (`argdiffs`).
"""

import functools

import jax.core as jc
import jax.tree_util as jtu
from jax import util as jax_util

from genjax._src.core.interpreters.forward import Environment, StatefulHandler
from genjax._src.core.interpreters.staging import stage
from genjax._src.core.pytree import Pytree
from genjax._src.core.typing import (
    Any,
    Callable,
    Generic,
    TypeVar,
)

R = TypeVar("R")

#######################################
# Change type lattice and propagation #
#######################################

###################
# Change tangents #
###################


class ChangeTangent(Pytree):
    pass


# These two classes are the bottom and top of the change lattice.
# Unknown change represents complete lack of information about
# the change to a value.
#
# No change represents complete information about the change to a value
# (namely, that it is has not changed).


@Pytree.dataclass
class _UnknownChange(ChangeTangent):
    pass


@Pytree.dataclass
class _NoChange(ChangeTangent):
    pass


UnknownChange = _UnknownChange()
NoChange = _NoChange()


#############################
# Diffs (generalized duals) #
#############################


@Pytree.dataclass(match_args=True)
class Diff(Generic[R], Pytree):
    """
    A class representing a difference type in incremental computation.

    This class pairs a value with a change tangent, which indicates whether the value
    has a known change. It is used to track changes in values during incremental computation.

    The Diff class stores a value along with a marker (ChangeTangent) saying whether or not
    that value has a known change.

    Create Diff instances with `Diff.no_change` and `Diff.unknown_change`.

    Note:
        Diff instances should only be used as leaves of an outer pytree. They should not contain nested Diff instances or be used as internal nodes in a pytree structure.

    Attributes:
        primal: The value being tracked.
        tangent: The change tangent indicating whether the value has a known change.
    """

    primal: R
    tangent: ChangeTangent

    def get_primal(self) -> R:
        return self.primal

    def get_tangent(self) -> ChangeTangent:
        return self.tangent

    #############
    # Utilities #
    #############

    @staticmethod
    def tree_diff(tree: R, tangent_tree: R) -> R:
        """
        Create a Diff tree by combining a primal tree with a tangent tree.

        This static method takes two trees of the same structure and combines them
        into a single tree where each node is a Diff instance. The primal values
        come from the `tree` argument, and the tangent values come from the
        `tangent_tree` argument.

        Args:
            tree: The tree containing primal values.
            tangent_tree: The tree containing ChangeTangent values, with the same
                          structure as `tree`.

        Returns:
            A new tree with the same structure as the input trees, where each
            node is a Diff instance combining the corresponding primal and
            tangent values.

        Note:
            The input trees must have the same structure, or a ValueError will
            be raised during the tree_map operation.
        """
        return jtu.tree_map(
            lambda p, t: Diff(p, t),
            tree,
            tangent_tree,
        )

    @staticmethod
    def no_change(tree: R) -> R:
        """
        Create a Diff tree with NoChange tangents for all nodes, used to represent a tree where no values have changed.

        Args:
            tree: The input tree to be converted.

        Returns:
            A new tree with the same structure as the input, where each
            node is a Diff instance with the original value as the primal
            and NoChange as the tangent.

        Note:
            This method first extracts the primal values from the input tree
            (in case it already contains Diff instances) before creating the
            new Diff tree. If any leaf in the input tree is already a Diff
            instance, its existing ChangeTangent will be replaced with NoChange.
        """
        primal_tree: R = Diff.tree_primal(tree)
        tangent_tree: R = jtu.tree_map(lambda _: NoChange, primal_tree)
        return Diff.tree_diff(primal_tree, tangent_tree)

    @staticmethod
    def unknown_change(tree: R) -> R:
        """
        Create a Diff tree with UnknownChange tangents for all nodes, used to represent a tree where values may have changed.

        Args:
            tree: The input tree to be converted.

        Returns:
            A new tree with the same structure as the input, where each
            node is a Diff instance with the original value as the primal
            and UnknownChange as the tangent.

        Note:
            This method first extracts the primal values from the input tree
            (in case it already contains Diff instances) before creating the
            new Diff tree. If any leaf in the input tree is already a Diff
            instance, its existing ChangeTangent will be replaced with UnknownChange.
        """
        primal_tree: R = Diff.tree_primal(tree)
        tangent_tree: R = jtu.tree_map(lambda _: UnknownChange, primal_tree)
        return Diff.tree_diff(primal_tree, tangent_tree)

    @staticmethod
    def tree_primal(v: R | "Diff[R]") -> R:
        """
        Converts a pytree that may contain Diff instances into a pytree of primal values.

        Args:
            v: A tree structure that may contain Diff instances or regular values.

        Returns:
            A new tree with the same structure as the input, where all Diff instances are replaced with their primal values.
        """

        def _inner(v) -> R:
            if isinstance(v, Diff):
                return v.get_primal()
            else:
                return v

        return jtu.tree_map(_inner, v, is_leaf=Diff.is_diff)

    @staticmethod
    def tree_tangent(v: "R | Diff[R]") -> R:
        """
        Converts a pytree that may contain Diff instances into a pytree of ChangeTangent values.

        Args:
            v: A tree structure that may contain Diff instances or regular values.

        Returns:
            A new tree with the same structure as the input, where all Diff instances are replaced with their ChangeTangent values, and all other values are replaced with UnknownChange.
        """

        def _inner(v: R | Diff[R]) -> ChangeTangent:
            if isinstance(v, Diff):
                return v.get_tangent()
            else:
                return NoChange

        return jtu.tree_map(_inner, v, is_leaf=Diff.is_diff)

    #################
    # Static checks #
    #################

    @staticmethod
    def is_diff(v: Any) -> bool:
        """
        Checks if a value is a Diff instance.

        Args:
            v: The value to check.

        Returns:
            True if the value is a Diff instance, False otherwise.
        """
        return isinstance(v, Diff)

    @staticmethod
    def is_change_tangent(v: Any) -> bool:
        """
        Checks if a value is a ChangeTangent instance.

        Args:
            v: The value to check.

        Returns:
            True if the value is a ChangeTangent instance, False otherwise.
        """
        return isinstance(v, ChangeTangent)

    @staticmethod
    def static_check_tree_diff(v) -> bool:
        """
        Returns true if all leaves in a pytree are Diff instances, False otherwise.
        """
        return all(
            map(
                Diff.is_diff,
                jtu.tree_leaves(v, is_leaf=Diff.is_diff),
            )
        )

    @staticmethod
    def static_check_no_change(v) -> bool:
        """
        Returns true if all leaves in a pytree are NoChange, False otherwise.
        """
        return all(
            map(
                lambda leaf: isinstance(leaf, _NoChange),
                jtu.tree_leaves(Diff.tree_tangent(v), is_leaf=Diff.is_change_tangent),
            )
        )


#################################
# Generalized tangent transform #
#################################


# TODO: currently, only supports our default lattice
# (`Change` and `NoChange`)
def default_propagation_rule(prim, *args, **_params):
    check = Diff.static_check_no_change(args)
    args = Diff.tree_primal(args)
    outval = prim.bind(*args, **_params)
    if check:
        return Diff.no_change(outval)
    else:
        return Diff.unknown_change(outval)


@Pytree.dataclass
class IncrementalInterpreter(Pytree):
    custom_rules: dict[jc.Primitive, Callable[..., Any]] = Pytree.static(
        default_factory=dict
    )

    def _eval_jaxpr_forward(
        self,
        _stateful_handler,
        jaxpr: jc.Jaxpr,
        consts: list[Any],
        primals: list[Any],
        tangents: list[ChangeTangent],
    ):
        dual_env = Environment()
        jax_util.safe_map(dual_env.write, jaxpr.constvars, Diff.no_change(consts))
        jax_util.safe_map(
            dual_env.write, jaxpr.invars, Diff.tree_diff(primals, tangents)
        )
        for _eqn in jaxpr.eqns:
            induals = jax_util.safe_map(dual_env.read, _eqn.invars)
            # TODO: why isn't this handled automatically by the environment,
            # especially the line above with _jaxpr.constvars?
            induals = [
                Diff(v, NoChange) if not isinstance(v, Diff) else v for v in induals
            ]
            subfuns, params = _eqn.primitive.get_bind_params(_eqn.params)
            args = subfuns + induals
            if _stateful_handler and _stateful_handler.handles(_eqn.primitive):
                outduals = _stateful_handler.dispatch(_eqn.primitive, *args, **params)
            else:
                outduals = default_propagation_rule(_eqn.primitive, *args, **params)
            if not _eqn.primitive.multiple_results:
                outduals = [outduals]
            jax_util.safe_map(dual_env.write, _eqn.outvars, outduals)

        return jax_util.safe_map(dual_env.read, jaxpr.outvars)

    def run_interpreter(self, _stateful_handler, fn, primals, tangents, **kwargs):
        def _inner(*args):
            return fn(*args, **kwargs)

        closed_jaxpr, (flat_primals, _, out_tree) = stage(_inner)(*primals)
        flat_tangents = jtu.tree_leaves(
            tangents, is_leaf=lambda v: isinstance(v, ChangeTangent)
        )
        jaxpr, consts = closed_jaxpr.jaxpr, closed_jaxpr.literals
        flat_out = self._eval_jaxpr_forward(
            _stateful_handler,
            jaxpr,
            consts,
            flat_primals,
            flat_tangents,
        )
        return jtu.tree_unflatten(out_tree(), flat_out)


def incremental(f: Callable[..., Any]):
    @functools.wraps(f)
    def wrapped(
        _stateful_handler: StatefulHandler | None,
        primals: tuple[Any, ...],
        tangents: tuple[Any, ...],
    ):
        interpreter = IncrementalInterpreter()
        return interpreter.run_interpreter(
            _stateful_handler,
            f,
            primals,
            tangents,
        )

    return wrapped



# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from genjax._src.generative_functions.static import (
    AddressReuse,
    StaticGenerativeFunction,
    StaticRequest,
    gen,
    trace,
    trace_p,
)

__all__ = [
    "AddressReuse",
    "StaticGenerativeFunction",
    "StaticRequest",
    "gen",
    "trace",
    "trace_p",
]
